/*                       -*- linux-c -*-

   WRIGHT.C a Linux modular char driver for a Wright Instruments CCD
   camera, model 2.

   Copyright (C) 1999 Michael Ashley, Joseph Lazar

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   You can contact the authors at m.ashley@unsw.edu.au
   (Dr. Michael Ashley, School of Physics, University of New South Wales,
   Sydney NSW 2052, Australia) and jlazar@xperts.hu (Xperts Software Ltd,
   Hungary).

   For the latest copy of the driver, visit 
   http://www.phys.unsw.edu.au/~mcba/apt.html

   This is a Linux modular char driver for the Wright Instruments 
   CCD camera model 2.

   You will need to edit this file to suit your camera (search for
   CCD_PHYSICAL below).    

   The driver has the following properties:
   ----------------------------------------

   - It provides autodetection of the CCD's IO base (yes) and IRQ
     number (not yet...), with the ability to explicitly set each, if
     desired, when the module is loaded.

   - It dynamically allocates its major device numbers. 

   - It fully implements the char operations
     read/write/select/open/release/fsync, and provides a minimal lseek.

   - read/write are provided with blocking and non-blocking interfaces.

   - It works with kernel versions 2.0, 2.2, and 2.4. It is strongly
     recommended to use kernel 2.4 since this is the system for which
     the driver is most thoroughly tested.

   Compile with:
   -------------

   cc -D__KERNEL__ -DMODULE -DLINUX -Dlinux -O2 -Wall 
       -Wstrict-prototypes -I/usr/include -c  wright.c -o wright.o

   Makefile:
   ---------

     INCLUDEDIR = /usr/include
     CC=gcc
     CFLAGS = -D__KERNEL__ -DMODULE -DLINUX -Dlinux -O2  -Wall -Wstrict-prototypes -I/usr/include
     # Extract version number from headers.
     VER = $(shell awk -F\" '/REL/ {print $$2}' $(INCLUDEDIR)/linux/version.h)
     OBJS = wright.o
     all:	$(OBJS)

   Compile-time defines:
   ---------------------

   NDEBUG    if not defined, the source includes additional "assert" tests.
             For the production code, you might like to define this,
             but take care...

   DEBUB_VERBOSE if defined, verbose debugging information is sent to the
             kernel log file.

   SIMULATOR if defined, software is used to simulate the operation of the
             CCD. This is useful for debugging without the hardware. Note
             that the driver still reserves an IRQ and ioports, so you need
             to ensure that the values used are free.
   
   Usage:

   To load the module, e.g.,

   /sbin/insmod wright.o
   /sbin/insmod wright.o io=0x360
   /sbin/insmod wright.o io=0x360 irq=15

   To unload the module:

   /sbin/rmmod wright

   To examine messages from the module:

   /bin/dmesg

   Once the driver is loaded, you can find the major device numbers
   for the CCD camera by using:

   grep ccd0 /proc/devices
   grep ccd1 /proc/devices

   To communicate with the driver you need to create at least one char 
   special file, e.g., if the ccd0 major device number is 253, use:

   mknod /dev/ccd0 c 253 0  ; creates the device special file

   To use the simulator:
   ---------------------

   (1) Find a free IRQ (use "cat /proc/interrupts" to identify the 
       ones that are used).

   (2) Verify that ioports 0x0300 to 0x0311 are free (use
       "cat /proc/ioports"). If these ports are not free, you will
       need to change CCD_HARDWARE_PORT in this file.

   (3) Edit the Makefile to have "-DSIMULATOR" in the CFLAGS.

   (4) make
       insmod wright.o io=0x300 irq=8 {do this as root}
       dmesg                          {check for any errors}
       cat /proc/ccd                  {should work}
       grep ccd0 /proc/devices        {find the major number}
       mknod -m 666 /dev/ccd0 c 253 0 {where 253 was the number found}

   (5) Edit the apt_image script and change $PATH_TO_EXECUTABLES 
       and $PATH_TO_IMAGES to, e.g., /tmp/images.

   (6) mkdir /tmp/images
       cd /tmp/images
       mkdir cache
       ./apt_image -o junk.fits -exp 1  {will take 1 second exposure}

       The resultant image should be 800x1200 pixels in size, with
       a ramp starting from the bottom left corner.
 
   Basic philosophy of the driver:
   -------------------------------

   The primary interface to the driver is through the device special
   file. You send commands to the camera by writing to the file, and
   you retrieve image data from the camera by reading the file. The
   commands that you send are in the form of single ASCII characters,
   with optional trailing numeric arguments. For example, 'g0' sets the
   camera gain, 'F' turns on the cooling fan, 'H' initiates the
   readout of a row from the CCD. For a complete list of commands,
   see the 'ccd_service()' and ccd_parameters()' routines below.

   The commands necessary to control the operation of the camera and
   readout an image can be placed into a text file, which can then 
   simply be written (using, e.g.,  'cat') to the device special file.  
   If at the same time you read (using 'cp') from the file, you will
   receive a FITS format image.

   NOTE: it is recommended that you use my program apt_image
         for sending commands to the driver.

   There are two device special files: ccd0 and ccd1. The first is
   used to send commands to the camera, and to read FITS images. The
   second file can be used as an alternative conduit for data from the
   camera; for example: you could direct a small array of pixels to ccd1
   for autoguiding purposes during an exposure.
 
   The commands that you send to the camera are checked carefully for
   reasonableness (e.g., that you have loaded the parameter memeory
   correctly before initiating readout). If there is an error, the
   driver writes a log message to the system log (use "dmesg" to
   read it), and then becomes inoperative. To recover from this
   situation you will need to terminate all processes that are talking to
   the driver, then unload/reload the module (using "rmmod/insmod").
   It should not be necessary to power-cycle the camera, since the
   driver should prevent inadvisable commands reaching the hardware.

   To create a FITS image:

   cp /dev/ccd0 image.fits & ; reads from the CCD into a FITS image file
   cat commands > /dev/ccd0  ; the file "commands" contains the commands needed
                                 to generate the FITS image

   The /proc filesystem interface:

   cat /proc/ccd          

   will display something like:

         Wright CCD device driver, version 010518a, status 0xffc0, 990951402
         Wait countdown = 123.23 secs
	 Output head-tail = 0; reschedules = 0
         The DMA buffer has become full 0 times
         dma residue = 0; status 0xffc0
         1 1 1 1 1 1 1 1 1 1 Fan is off, Cooler is on
         Shutter is closed
         Horizontal operation is not in progress
         Vertical operation is not in progress
         Power supply is OK
         Power supply is OK
         Last command 0x301; sequence #34127
         C     g0 s0               z00000 
         p00 | m0 r0000 d0000 - m4 r0000 b0000
         p01 | m3 r0001 d0008 - m6 r0800 b0000
         p03 | m0 r0000 d0000 - m4 r0800 b0000
         p04 | m3 r2400 d0010 - m5 r0000 b0000
         p05 | m3 r0000 d0096 - m5 r0000 b0000

   You can also look in:
           
   /proc/ioports        ; to see the io address used by the CCD
   /proc/interrupts     ; to see the IRQ number being used, and the
                          number of interrupts that the CCD 
                          has generated
   /proc/devices        ; to see the major device numbers
   /proc/modules        ; to see that the module is loaded


   Notes for anyone who wants to modify this file:
   -----------------------------------------------

   In order to ensure that this module is as bug-free as
   possible, it is important to keep the flow of control from the
   user to the hardware as logical as possible. This is achieved
   by following the following guidelines:

   The ONLY routines that access the hardware directly are:

     ccd_send_command
     ccd_receive_word
     ccd_read_status
     ccd_find_hardware (during module initialisation only)

   These routines are ONLY called externally by

     ccd_process_command

   (Actually, the above is violated by ccd_ioctl for the test function,
    and by ccd_find_hardware).
   
   And this routine in turn is ONLY called externally by

     ccd_service

   Which in turn is ONLY called from a task queue, and it can
   ONLY be placed on a task queue by 

     ccd_queue_service_task 

   Which in turn is ONLY called by one of:

     ccd_isr                 (the CCD interrupt)
     ccd_timer_service_task  (a programmed wait)
     ccd_service             (i.e., the task can re-queue itself) 
     ccd_read0               (if the output buffer has been read)
     ccd_read1               (if the output buffer has been read)
     ccd_write               (if new data is available)


  TO DO:
  ------
 
   - what happens for delays less than one tick (hopefully the timer
     will go off immediately)?
   - add GAIN and NOISE to header
   - use minor device numbers rather than two majors
   - should check the return value of schedule_task();
   - procmem if more than 512 characters? (see pcwd for a neat way
     of handling this).
   - difference between proc_info and proc_read?
   - should load on demand
   - should set the value of p->delay sensibly when waiting for
     horizontal/vertical transfers, so as to minimise time wasted. 
   - check to ensure we can recover if DMA is aborted
   - if an assertion fails, need to disconnect all processes that use the
     driver, and ensure the usage count is zero, so that the module
     can be unloaded.
   - need a "alive" checking routine, scheduled every ten seconds or so.
     It might send a signal to a process to indicate failure.
   - implement "delay" command, e.g., for delaying by a few millisecs after
     the shutter is closed.
   - the receive and send timeouts should take into account "loops_per_second"
   - As a last-ditch attempt to avoid ring_full, could wake-up read/writers
     before a byte is lost...
   - Hm... shouldn't really allow two processes to read/write at once.
   - check select operation for blocking
   - check DANGER
   - attribute sources (e.g., kernel book, Wright manual)


   HARDWARE INFORMATION:

   The communication link runs at 10 Mbps along a 50-ohm coaxial
   cable. Each end is terminated in an isolation transformer (type "601"
   or "73601/3"). The signal reaches +/- 2 VDC.

   A data frame lasts 2.5 microseconds and contains 17 bits of
   Manchester-encoded data, and appears to have a start bit, a few bits
   of preamble for synching, and a stop bit.

   Data frames are produced every 5.4 microseconds when using the fast
   digitising mode.

    MODIFICATION HISTORY

    M0004 - by Michael Ashley, (mcba), April 2001.

            Ported to kernel 2.4:
               - VALID_PAGE was changed to VALID__PAGE, since the earlier
                 name is now used by the kernel.
               - queue_task no longer supports tq_scheduler, so
                 schedule_task is used instead.
               - The Wait Queue system interface has changed in 2.4.
               - The fsync file operation has a new parameter "datasync".
               - The layout of the file_operations struct has changed.
               - The layout of the proc_dir_entry struct has changed.
               - The API proc_register has been replaced by 
                 create_proc_read_entry?
                 create_proc_info_entry?; proc_unregister by remove_proc_entry.
                 must not have an initialiser for static proc_dir_entry

            The software simulator now works! Simply compile with the
            -DSIMULATOR option, and you can use the module without needing
            a CCD camera attached. The image is initialised to a ramp from
            the bottom left corner. Note the "=" command, which can be used
            to initialise the simulated image.

            Removed the ability of the 'y' and 'Y' commands to read out
            multiple rows - this was a cause of several subtle bugs.

            The proc entry now gives you an exposure countdown timer. This
            allows you to construct simple X11 programs to display the
            remaining exposure time.

            Exposure timing now much more accurate (10s of microseconds),
            and exposures can be started at precise UTC times (using the
            new 'M' command). 'W' can be used to insert precision delays
            for various purposes (e.g., multiple exposures without reading
            out the chip).

            Added a test mode accessible via ioctl. This allows a series
            of status requests to be sent to the CCD, and you can then 
            monitor the communication with an oscilloscope.

            Cleaned up the flow of control in the module, hopefully removing
            some subtle bugs.

	    Added some spin_lock protection, but this is still rudimentary.

            Removed a lot of the exposure control stuff added in M0001
            since the code was not safe (e.g., it could cause recursive
            calls of routine that should not be called recursively).
            Unfortunately, this removes the ability to open/close the
            shutter during an exposure from ioctl. This is not a trivial
            problem to solve without introducing bugs.

            Added some additional initialisation of the ISA card when the
            module is first loaded.

            Made lots of other changes, which aren't documented at the
            moment due to pressure of work.

            Removed CCD_NEXT_LINE since this had bugs, and is not
            necessary.

    M0003 - by Michael Ashley, (mcba), January 2000.

            ccd_read_status change.

    M0002 - by Michael Ashley, (mcba), November 1999.

            Using a 32-bit integer for delay, so that we can have
            exposure times greater than 2^16 milliseconds.

	    Added checks for the 'O' and 'D' commands to ensure that
            the exposure time is greater than zero.

            Increased size of FITS strings from 40 to 60 chars.

            Made DATE-OBS use a 4-digit year. Note that there are some
            Y2K problems with the "tm" structure and its use.

            Format change to TIME-BEG and TIME-END so that the
            fractional seconds was correct if < 0.1.

            FITS description of the TIME-BEG and TIME-END fields
            varies between UTC and local, depending on the timezone.

            Added information message if we fail to allocate a
            DMA buffer on startup.

            Reinserted definition of *ccd_exq, to allow compilation
            with a 2.2 kernel

            Minor cosmetic changes.

            Removed some vestigial code.

    M0001 - by Joseph Lazar, (jol)
            Added ccd_status to record different kind of activities
            we have initiated.
            - SHUTTER_TIMER_IS_ON - timer started
            - SHUTTER_IS_OPEN - is open
            - SHUTTER_IN_UNKNOWN_STATE - initial state at module load time
            - SHUTTER_HANGUP - we are in a hangup situation

            Added new members to "Expdata" to hold start
            time for a given exposure. They are filled out in
            ccd_process_command() and ccd_close_shutter().

            Add an 'unloadable' wait_queue var to allow waiting in the
            cleanup_module() in case an ISR is still in the queue. (To allow
            return in any case pull a timer for it.)

            Enable usage of io=, irq=, dma= at insmod.
            
            Introduce __LITTLE_ENDIAN to do swapping only on PC like
            achitechtures.

            Corrected loop in ccd_service() to use "loops_per_second".

            Added 'I' command to allow FITS header parameter settings.

            Added 'D' command to allow dark expos.

            Added ABORT, CANCEL, HANGUP and RESUME IOCTL commands and
            full handling of it.

            Eliminated a couple of DANGER stuff.

            Added ifndef NDEBUG for dma_*time variable usages.

            Added cooler status report to /proc infos and also code to
            init 'ccd_init.cooler'.

            Some infos in the /proc can be old, because the status word
            is read only after some commands. Now we try to get it at
            the time when the user makes a 'cat', thus giving acurate
            status. If we can't get it we inform the user that the camera
            might be turned off.

            'last_command' is saved only when it is not a status return
            thus giving the possibility to see the last 'real' command
            sent to the device.  */

// MACROS for programming constructs -----------------------------------------

#define VERSION_CODE(vers,rel,seq) (((vers)<<16) | ((rel)<<8) | (seq))

// Here is an implementation of "assert" for modules. It does not
// abort the program, like a real one should, but does the following on
// a failed assertion:
//
//      (1) Prints a message in the system log. You should use "dmesg" 
//          occasionally to check for these.
//
//      (2) Sets a global variable "assertion_has_failed", which is then
//          used to effectively disable the module.
//
//      (3) Performs an arbitrary user defined "action" (usually
//          something like "return -EINVAL"). If you want no action,
//          use an argument of ((void) 0).

#ifdef NDEBUG
#define assert(expression,action) ((void) 0)
#else
#define assert(expression,action) \
    if (!(expression)) { \
      printk (KERN_ALERT "ASSERTION FAILED: %s, %s at line %i in %s\n", \
              #expression, __FILE__, __LINE__, __PRETTY_FUNCTION__); \
      assertion_has_failed = 1; \
      action; \
    }
#endif

#define RETURN_ON_FAILURE(x) if (0 != (ret = (x))) return ret
#define MIN(x,y) ((x) < (y) ? (x) : (y));

// "#include" FILES ----------------------------------------------------------

#include <linux/version.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/types.h>

#ifndef SIMULATOR
#include <asm/io.h>
#include <asm/dma.h>
#endif				/* not SIMULATOR */

#include <asm/irq.h>

#if LINUX_VERSION_CODE >= VERSION_CODE(2,1,23)
#include <linux/poll.h>
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,0,0)
#error This module will not work with kernels older than 2.0.0.
#endif

#if LINUX_VERSION_CODE >= VERSION_CODE(2,5,0)
#error This module requires modification for kernels newer than 2.4.
#endif

// OPERATING SYSTEM DATA TYPES THAT CHANGE BETWEEN KERNELS -------------------

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
#define lseek_t      int
#define read_write_t int
#define count_t      int
#define release_t    void
#define release_return(value) return
#else
#define lseek_t      loff_t
#define read_write_t ssize_t
#define count_t      long
#define release_t    int
#define release_return(value) return(value)
#endif

// TEMPORARY JUNK USED DURING PROGRAM DEVELOPMENT ----------------------------

#ifndef NDEBUG
unsigned long dma_starting_time, dma_delta_time, dma_times[10];
unsigned int last_command = 0;

unsigned long last_command_number = 0;
int number_of_reschedules = 0;
#endif

// GLOBAL VARIABLES  ---------------------------------------------------------

static int assertion_has_failed = 0;
static int module_shutting_down = 0;
static int throw_away_dma_data  = 0;
static struct timeval wait_timeval;

unsigned long dma_begin[2], dma_end[2], dma_read_pointer[2],
    dma_next_read_pointer[2], dma_next_write_pointer[2],
    dma_write_pointer[2], dma_last_written_pointer[2];

struct tq_struct ccd_task;

static char *ccd_name = "Wright CCD";

static char *ccd_version = "011015a";	// Format: YYMMDDx

static int io = 0;		// The IO base address for the CCD card.
static int irq = 0;		// The IRQ line used by the CCD card.
static int dma = 0;		// The DMA channel used by the CCD card.

#if LINUX_VERSION_CODE > VERSION_CODE(2,1,0)
MODULE_PARM(io, "i");
MODULE_PARM(irq, "i");
MODULE_PARM(dma, "i");
#endif

// STRUCT DEFINITIONS -----------------------------------------------------

struct ccd_command_struct {
    u8 port;
    u32 delay;
    u16 outword;
};

// enums THAT THE USER MAY WANT TO CHANGE -------------------------------

enum {
    MAX_WORDS_PER_DMA_TRANSFER = 1200,
    CCD_RING_O_BUFFER_LENGTH   = 256 * sizeof(struct ccd_command_struct),
    CCD_WORDS_BEFORE_YIELDING  = 2560
};

// enums FOR IMMUTABLE SOFTWARE USE -------------------------------------

enum {
    CCD_STOP_HORIZONTAL_TRANSFERS     = 0x10,
    CCD_WAIT_FOR_HORIZONTAL_TRANSFERS = 0x11,
    CCD_STOP_VERTICAL_TRANSFERS       = 0x12,
    CCD_WAIT_FOR_VERTICAL_TRANSFERS   = 0x13
};

// enums FOR PROPERTIES OF THE CCD CAMERA -------------------------------

enum {
    CCD_PHYSICAL_XDIM = 770,	/* The X dimension of the CCD chip    */
    CCD_PHYSICAL_YDIM = 1152,	/* The Y dimension of the CCD chip    */
    CCD_FIFO_SIZE = 2048,	/* The controller fifo length         */
    CCD_READOUT_REGISTER_LENGTH = 814, /* The length of the readout register */
    CCD_READOUT_REGISTER_OFFSET = 22,  /* The number of leading blank pixels */
    CCD_IRQ_NUMBER = 11,	/* The Wright camera interrupt number */
    CCD_DMA_CHANNEL = 5,	/* The Wright camera DMA channel      */


    CCD_IO_EXTENT = 18,		/* The range in bytes of IO space 
				   used by the CCD ISA card           */
    CCD_RECEIVE = 0x80,		/* Ready to receive bit               */
    CCD_SEND = 0x20,		/* Ready to send bit                  */
    CCD_RECEIVE_TIMEOUT = 1000,	/* DANGER Receive timeout             */
    CCD_SEND_TIMEOUT = 1000,	/* DANGER Send timeout                */
    COMMAND_MASK = 0x0f00,
    COMMAND_SELECT_REGISTER = 0x0000,
    COMMAND_WRITE_REGISTER = 0x0100,
    COMMAND_VERTICAL_TRIGGER = 0x0200,
    COMMAND_HORIZONTAL_TRIGGER = 0x0300,
    COMMAND_ANALOG_CONTROL = 0x0400,
    COMMAND_SHUTTER = 0x0800,
    COMMAND_COOLER = 0x0a00,
    COMMAND_STATUS_RETURN = 0x0b00,
    COMMAND_STATUS_RESET = 0x0c00,
    PARAMETER_VERTICAL_MODE = 0,
    PARAMETER_VERTICAL_REPEAT_COUNT_LO = 2,
    PARAMETER_VERTICAL_REPEAT_COUNT_HI = 3,
    PARAMETER_VERTICAL_DELAY_LO = 4,
    PARAMETER_VERTICAL_DELAY_HI = 5,
    PARAMETER_HORIZONTAL_MODE = 16,
    PARAMETER_HORIZONTAL_REPEAT_COUNT_LO = 18,
    PARAMETER_HORIZONTAL_REPEAT_COUNT_HI = 19,
    PARAMETER_HORIZONTAL_BINNING_LO = 20,
    PARAMETER_HORIZONTAL_BINNING_HI = 21,
    BIT_PAGE = 0x0080,
    BIT_SHIFT = 0x0080,
    BIT_HORIZONTAL = 0x0080,
    BIT_VERTICAL = 0x0100
};

// HELPFUL MACROS TO CLARIFY INTENT (HOPEFULLY) -------------------------

#define CCD_RING_O_EMPTY \
  (ccd_ring_o_head == ccd_ring_o_tail)
#define CCD_RING_O_FULL \
  (1 == (ccd_ring_o_tail + CCD_RING_O_BUFFER_LENGTH - ccd_ring_o_head) % \
   CCD_RING_O_BUFFER_LENGTH)

#define VALID__PAGE       (ccd_mode.page < 0x40)
#define VALID_REG         (ccd_mode.reg  < 0x20)
#define CURRENT_PAGE_PAR   ccd_mode.par[ccd_mode.page]
#define CURRENT_PAGE_INIT  ccd_init.par[ccd_mode.page]

#define CCD_IS_READY_TO_RECEIVE (inw_p (CCD_STATUS_PORT) & CCD_RECEIVE)
#define CCD_IS_SENDING          (inw (CCD_STATUS_PORT) & CCD_SEND)
// slight DANGER: not using inw_p above

#define READY_FOR_HORIZONTAL_TRIGGER \
  (VALID__PAGE && CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_MODE] && \
   CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_REPEAT_COUNT_LO] && \
   CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_REPEAT_COUNT_HI] && \
   CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_BINNING_LO] && \
   CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_BINNING_HI])

#define DIGITIZING \
  (VALID__PAGE && CURRENT_PAGE_INIT[PARAMETER_HORIZONTAL_MODE] && \
   (CURRENT_PAGE_PAR[PARAMETER_HORIZONTAL_MODE] & 0x0002))

#define DMA_ENABLED \
  (ccd_init.dma && ccd_mode.dma == 0x0001)

#define READY_FOR_VERTICAL_TRIGGER \
  (VALID__PAGE && CURRENT_PAGE_INIT[PARAMETER_VERTICAL_MODE] && \
   CURRENT_PAGE_INIT[PARAMETER_VERTICAL_REPEAT_COUNT_LO] && \
   CURRENT_PAGE_INIT[PARAMETER_VERTICAL_REPEAT_COUNT_HI] && \
   CURRENT_PAGE_INIT[PARAMETER_VERTICAL_DELAY_LO] && \
   CURRENT_PAGE_INIT[PARAMETER_VERTICAL_DELAY_HI])

// "#defines" FOR IMMUTABLE CCD PROPERTIES ----------------------------------

#define CCD_DATA_PORT           (io)	// The CCD's IO base address
#define CCD_STATUS_PORT         (io +  2)	// Bus status byte
#define CCD_DMA_PORT            (io +  4)	// DMA control
#define CCD_CAM_INT_PORT        (io +  6)	// Camera interrupt control
#define CCD_CANCEL_CAM_INT_PORT (io +  8)	// Cancel camera interrupt request
#define CCD_PLL_PORT            (io + 10)	// Parallel IO port
#define CCD_PLL_INT_PORT        (io + 12)	// Parallel interrupt control
#define CCD_CANCEL_PLL_INT_PORT (io + 14)	// Cancel pll interrupt request
#define CCD_OFFSET_PORT         (io + 16)	// Set data offset


#define SECS_TO_WAIT_BEFORE_CLEANUP 2

#include    <asm/byteorder.h>	// For the sake of htons()
struct tm {
    int tm_sec;			/* Seconds. [0-60] (1 leap second) */
    int tm_min;			/* Minutes. [0-59] */
    int tm_hour;		/* Hours.   [0-23] */
    int tm_mday;		/* Day.     [1-31] */
    int tm_mon;			/* Month.   [0-11] */
    int tm_year;		/* Year - 1900.  */
    int tm_wday;		/* Day of week. [0-6] */
    int tm_yday;		/* Days in year.[0-365] */
    int tm_isdst;		/* DST.     [-1/0/1] */
};
void offtime(time_t * t, long int offset, struct tm *tp);

//-- M0001 -------------------------------------------------------------------

char header[72][80];
struct {
    long time;			// Real expo time in milliseconds - used to nearest 1ms
    long time_todo;		// Exposure time in milliseconds - used to nearest 1ms
    short gain;			// Camera gain 0=low 1=high
    short shift;		// Shift data shift bits to left during readout
    char zeroa;			// Adjust zero level 0=no 1=yes
    short zero;			// Zero level to adjust true zero to
    char speed;			// Readout speed 0=slow 1=fast
    short x;			// Size of chip in screen x direction (usually longest)
    short y;			// Size of chip in screen y direction (usually shorter)
    short x1;			// 1st x pixel to read out (1<=x1<=x)
    short y1;			// 1st y pixel to read out (1<=y1<=y)
    short x2;			// last x pixel to read out (x1<=x2<=x)
    short y2;			// last y pixel to read out (y1<=y2<=y)
    short xb;			// Binning factor on x axis (adjustable by software)
    short yb;			// Binning factor on y axis (adjustable by software)

    // Order, size, and number of the following  MUST stay, because at 
    // reset we just memset them. Each must be the same size as obs[].

    char imagetyp[60];		// "Iim"
    char observer[60];		// "Ibs"
    char object[60];		// "Ij"
    char ra[60];		// "Ira"
    char dec[60];		// "Id"
    char subset[60];		// "Isu"
    char inst[60];		// "Iin"
    char instrum[60];		// "Iu"
    char lat[60];		// "Il"
    char longi[60];		// "In"
    char alt[60];		// "Ia"
    char email[60];		// "Iem"
    char obs[60];		// "Io"
    char xpix[60];		// "Ix"
    char ypix[60];		// "Iy"
    char psc[60];		// "Ip"
    char tel[60];		// "It"
    char www[60];		// "Iw"
    char fits[480];		// FITS lines (max 6)

    int flines;			// # of FITS header lines added via "If"
    int ctype;			// "Ic"
    int tz;			// "Iz"
    struct timeval exp_start_time;	// Exposure start time
    struct timeval exp_end_time;	// Exposure end time
} Expdata;

void ccd_write_fits_header(void)
{
    int i, j, naxis1, naxis2;
    int secs, usec, ifits;	// M0001
    char buf[60], *p;		// M0001
    struct tm exptime;

    // M0001 - Set by 'I' commands, if not then use defaults
    if (!Expdata.x1)
	Expdata.x1 = 1;
    if (!Expdata.x2)
	Expdata.x2 = 800;
    if (!Expdata.xb)
	Expdata.xb = 1;
    if (!Expdata.y1)
	Expdata.y1 = 1;
    if (!Expdata.y2)
	Expdata.y2 = 1200;
    if (!Expdata.yb)
	Expdata.yb = 1;

    naxis1 = (Expdata.x2 - Expdata.x1 + Expdata.xb) / Expdata.xb;
    naxis2 = (Expdata.y2 - Expdata.y1 + Expdata.yb) / Expdata.yb;
    i = 0;
    sprintf(header[i++], "SIMPLE  =                    T");
    sprintf(header[i++],
	    "BITPIX  =                   16 / signed 16-bit integers");
    sprintf(header[i++],
	    "NAXIS   =                    2 / 2-dimensional image");
    sprintf(header[i++], "NAXIS1  =           %10d / x dimension", naxis1);
    sprintf(header[i++], "NAXIS2  =           %10d / y dimension", naxis2);
    sprintf(header[i++],
	    "BSCALE  =                  1.0 / REAL = FILE * BSCALE + BZERO");
    sprintf(header[i++], "BZERO   =             +32768.0");
    sprintf(header[i++], "BAXIS1  =           %10d / x binning factor",
	    Expdata.xb);
    sprintf(header[i++], "BAXIS2  =           %10d / y binning factor",
	    Expdata.yb);
    sprintf(buf, "[%d:%d,%d:%d]", Expdata.x1, Expdata.x2, Expdata.y1,
	    Expdata.y2);
    sprintf(header[i++],
	    "DATASEC = '%-18.18s' / image section containing data", buf);
    sprintf(header[i++],
	    "VERSION = '%-18.18s' / version of the device driver used",
	    ccd_version);
    if (*Expdata.imagetyp)
	sprintf(header[i++], "IMAGETYP= '%-18.18s' / type of image",
		Expdata.imagetyp);
    if (*Expdata.subset) {
	sprintf(header[i++], "SUBSET  = '%-18.18s' / filter used",
		Expdata.subset);
	// M0001 - For the sake of IRAF
	sprintf(header[i++], "FILTERS = '%-18.18s' / filter used",
		Expdata.subset);
    }
    if (*Expdata.object)
	sprintf(header[i++],
		"OBJECT  = '%-18.18s' / the name of the object observed",
		Expdata.object);
    if (Expdata.ra)
	sprintf(header[i++], "RA      = %20.20s / right ascension [hours]",
		Expdata.ra);
    if (Expdata.dec) {
	sprintf(header[i++], "DEC     = %20.20s / declination [degrees]",
		Expdata.dec);
	sprintf(header[i++],
		"EQUINOX = 'J2000'              / equinox of coordinates");
    }
    if (Expdata.exp_start_time.tv_sec) {
	offtime((time_t *) & (Expdata.exp_start_time.tv_sec),
		(long) (Expdata.tz * 3600), &exptime);
	sprintf(header[i++],
		"DATE-OBS= '%02d/%02d/%4d'         / date of exposure [DD/MM/YYYY]",
		exptime.tm_mday, exptime.tm_mon + 1,
		1900 + exptime.tm_year);
	sprintf(header[i++],
		"TIME-BEG= '%02d:%02d:%02d.%06ld'    / %s time of exposure [HH:MM:SS.SSS]",
		exptime.tm_hour, exptime.tm_min, exptime.tm_sec,
		Expdata.exp_start_time.tv_usec,
		Expdata.tz ? "local" : "UTC");
	if (Expdata.exp_end_time.tv_sec) {
	    offtime((time_t *) & (Expdata.exp_end_time.tv_sec),
		    (long) (Expdata.tz * 3600), &exptime);
	    sprintf(header[i++],
		    "TIME-END= '%02d:%02d:%02d.%06ld'    / %s time of exposure end [HH:MM:SS.SSS]",
		    exptime.tm_hour, exptime.tm_min, exptime.tm_sec,
		    Expdata.exp_end_time.tv_usec,
		    Expdata.tz ? "local" : "UTC");
	    secs = Expdata.exp_end_time.tv_sec - Expdata.exp_start_time.tv_sec;
            usec = Expdata.exp_end_time.tv_usec - Expdata.exp_start_time.tv_usec;
            if (usec < 0) {
                usec += 1000000;
                secs--;
            }
	    sprintf(buf, "%d.%03d", secs, usec/1000);
	    sprintf(header[i++], "EXPTIME = %20.20s / exposure time [secs]",
		    buf);
	}
    }

    sprintf(header[i++],
	    "CAMSPEED= %20d / readout speed [0 = low; 1 = high]",
	    Expdata.speed);
    if (*Expdata.observer)
	sprintf(header[i++],
		"OBSERVER= '%-18.18s' / the name of the observer",
		Expdata.observer);
    if (Expdata.xpix)
	sprintf(header[i++], "XPIXELSZ= %20.20s / x pixel size [microns]",
		Expdata.xpix);
    if (Expdata.ypix)
	sprintf(header[i++], "YPIXELSZ= %20.20s / y pixel size [microns]",
		Expdata.ypix);
    if (Expdata.psc)
	sprintf(header[i++],
		"PLATESCL= %20.20s / plate scale [arcsecs/mm]",
		Expdata.psc);
    if (*Expdata.obs) {
	sprintf(header[i++], "OBSERVAT= '%-18.18s' / observatory",
		Expdata.obs);
	// M0001 - For the sake of IRAF
	sprintf(header[i++], "OBSVAT  = '%-18.18s' / observatory",
		Expdata.obs);
    }
    if (Expdata.lat)
	sprintf(header[i++],
		"LAT_OBS = %20.20s / observatory latitude [degrees]",
		Expdata.lat);
    if (Expdata.longi)
	sprintf(header[i++],
		"LONG_OBS= %20.20s / observatory longitude [degrees]",
		Expdata.longi);
    if (Expdata.alt)
	sprintf(header[i++],
		"ALT_OBS = %20.20s / observatory altitude [meters]",
		Expdata.alt);
    sprintf(header[i++], "TIMEZONE= %20d / observatory timezone [hours]",
	    Expdata.tz);
    if (*Expdata.tel)
	sprintf(header[i++], "TELESCOP= %-60.60s", Expdata.tel);
    if (*Expdata.instrum)
	sprintf(header[i++], "INSTRUME= %-60.60s", Expdata.instrum);
    if (*Expdata.inst)
	sprintf(header[i++], "INSTITUT= %-60.60s", Expdata.inst);
    if (*Expdata.email)
	sprintf(header[i++], "EMAIL   = %-60.60s", Expdata.email);
    if (*Expdata.www)
	sprintf(header[i++], "WWW     = %-60.60s", Expdata.www);

    // M0001 - Now add FITS lines that came thru 'If' commands
    for (ifits = 0; ifits < Expdata.flines; ifits++) {
	p = &Expdata.fits[ifits * 80];
	if (!p)
	    break;
	strncpy(header[i++], p, strlen(p));
    }

    // Zero the remaining lines of the header.

    i++;
    for (; i < 70; i++) {
	header[i++][0] = 0;
    }

    // Add the END record.

    sprintf(header[71], "END");

    // Convert all the NUL bytes to spaces.

    for (i = 0; i < 72; i++) {
	for (j = strlen(header[i]); j < 80; j++) {
	    header[i][j] = ' ';
	}
    }

}

//  GLOBAL VARIABLES  ------------------------------------------------------

static int ccd_possible_port_bases[] = { 0x360, 0x340, 0x320, 0x300, 0 };
				// These are the addresses that will be
				// probed if autodetection is used.
				// WARNING: the above array must have at 
				// least two elements AND the last one 
				// must be zero!

static unsigned int ccd_major0 = 0;
				// The major device number for the CCD;
				// you can specify a number, or use zero
				// to use dynamic allocation.

static unsigned int ccd_major1 = 0;
				// The major device number for the CCD;
				// you can specify a number, or use zero
				// to use dynamic allocation.

static unsigned int ccd_number;

static int servicing = 0;	// True if executing ccd_service().

static unsigned char ccd_ring_o_buffer[CCD_RING_O_BUFFER_LENGTH];
static unsigned long ccd_ring_o_head = 0;
static unsigned long ccd_ring_o_tail = 0;
static unsigned long ccd_ring_o_tail_save = 0;

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
static struct wait_queue
    *cleanup_queue = NULL,
    *ccd_inq0 = NULL, *ccd_inq1 = NULL, *ccd_outq = NULL, *ccd_exq = NULL;
#else
static wait_queue_head_t cleanup_queue, ccd_inq0, ccd_inq1, ccd_outq,
    ccd_exq;
#endif

static int have_requested_region = 0;
static int have_registered_device0 = 0;
static int have_registered_device1 = 0;
static int have_initialized_ccd = 0;
static int have_irq = 0;
static int have_dma = 0;
static int ccd_setup_to_use_interrupts = 0;

struct fasync_struct *ccd_async_queue;

// The following structure is used to hold a snapshot of the state of the
// CCD camera (in the global variable "ccd_mode"). "ccd_init" is used to
// keep track of which of the locations within "ccd_mode" have actually 
// been written to (e.g., if "ccd_init.page" is false, this means that
// "ccd_mode.page" has not been written to since the driver was initialised).

static struct ccd_mode_struct {
    u8 page;
    u8 reg;
    u8 shutter;
    u8 gain;
    u8 shift;
    u8 cooler;
    u8 dma;
    u8 dma_slot;
    u8 dma_in_progress;
    u8 readout_in_progress;
    u8 using_dma;
    u8 closing_file;
    u8 waiting;
    u8 stm;
    u8 cam_int;
    u8 direction;
    u8 fits_index;
    u16 pll;
    u16 pll_int;
    u16 offset;
    u16 status;
    u16 interface_status;
    u16 words_expected;
    u16 lines;
    u16 need_trigger;
    u16 dma_overwrite;
    struct timeval time;
    u8 par[64][32];
} ccd_mode = {

0xff, 0xff,}, ccd_init;

// FUNCTION PROTOTYPES ----------------------------------------------------

int ccd_process_command(struct ccd_command_struct *p);
void ccd_queue_service_task(int immediate);

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
lseek_t ccd_llseek(struct inode *inode, struct file *filp,
		   off_t off, int whence);
read_write_t ccd_read0(struct inode *inode, struct file *filp,
		       char *buf, count_t count);
read_write_t ccd_read1(struct inode *inode, struct file *filp,
		       char *buf, count_t count);
read_write_t ccd_write(struct inode *inode, struct file *filp,
		       const char *buf, count_t count);
#else
lseek_t ccd_llseek(struct file *filp, loff_t off, int whence);
read_write_t ccd_read0(struct file *filp,
		       char *buf, size_t count, loff_t * ppos);
read_write_t ccd_read1(struct file *filp,
		       char *buf, size_t count, loff_t * ppos);
read_write_t ccd_write(struct file *filp,
		       const char *buf, size_t count, loff_t * ppos);
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,23)
int ccd_select0(struct inode *inode, struct file *filp,
		int mode, select_table * table);
int ccd_select1(struct inode *inode, struct file *filp,
		int mode, select_table * table);
#else
unsigned int ccd_poll0(struct file *filp, poll_table * table);
unsigned int ccd_poll1(struct file *filp, poll_table * table);
#endif

int ccd_ioctl(struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg);
int ccd_open(struct inode *inode, struct file *filp);
release_t ccd_release(struct inode *inode, struct file *filp);

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
int ccd_fsync(struct inode *inode, struct file *filp);
int ccd_fasync(struct inode *inode, struct file *filp, int mode);
#elif LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
int ccd_fsync(struct file *filp, struct dentry *d);
int ccd_fasync(int fd, struct file *filp, int mode);
#else
int ccd_fsync(struct file *filp, struct dentry *d, int datasync);
int ccd_fasync(int fd, struct file *filp, int mode);
#endif

int ccd_send_line(unsigned char *line);

struct file_operations ccd_fops0 = {
#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,0)
    owner:   THIS_MODULE,
#endif
    llseek:  ccd_llseek,
    read:    ccd_read0,
    write:   ccd_write,
    poll:    ccd_poll0,
    ioctl:   ccd_ioctl,
    open:    ccd_open,
    release: ccd_release,
    fsync:   ccd_fsync,
    fasync:  ccd_fasync,
};

struct file_operations ccd_fops1 = {
#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,0)
    owner:   THIS_MODULE,
#endif
    llseek:  ccd_llseek,
    read:    ccd_read1,
    write:   ccd_write,
    poll:    ccd_poll1,
    ioctl:   ccd_ioctl,
    open:    ccd_open,
    release: ccd_release,
    fsync:   ccd_fsync,
    fasync:  ccd_fasync,
};

#ifdef SIMULATOR

// CCD SIMULATOR ------------------------------------------------------------

/* 
  The simulator works by taking over the "inw_p()", "in_p()" and "outw_p ()"
  system calls, and giving responses that mimic, to a some extent, the
  Wright Instruments hardware.
*/

#define inw    my_inw_p
#define inw_p  inw
#define outw_p my_outw_p

#define CCD_SATURATION 9999999
#define CCD_HARDWARE_PORT 0x300

static struct ccd_mode_struct simd, simi;

static u16 ccd_chip[CCD_PHYSICAL_YDIM][CCD_PHYSICAL_XDIM];
static u32 ccd_readout[CCD_READOUT_REGISTER_LENGTH];
static u32 ccd_output;
static u16 ccd_fifo[CCD_FIFO_SIZE];
static int ccd_fifo_in, ccd_fifo_out;

unsigned long in_last_time, in_this_time, in_delta_time;
unsigned long out_last_time, out_this_time, out_delta_time;

/* Defines and functions to replace dma.h */

static int dma_enabled = 0;
static u16 *dma_address = NULL;
static u16 dma_count = 0;
static u16 dma_residue = 0;

#define DMA_MODE_READ 0

void disable_dma(int channel)
{
    dma_enabled = 0;
    dma_address = NULL;
    dma_count = 0;
    dma_residue = 0;
}

void clear_dma_ff(int channel)
{
}
void set_dma_mode(int channel, int mode)
{
}

void enable_dma(int channel)
{
    assert((dma_address != NULL) &&
	   (dma_count == 0) && (dma_residue != 0), (void) 0);
    dma_enabled = 1;
}

void free_dma(int channel)
{
    disable_dma(channel);
}

void *virt_to_bus(void *address)
{
    return address;
}

int get_dma_residue(int channel)
{
    return dma_residue;
}

int request_dma(int channel, char *device)
{
    return 0;
}

void set_dma_addr(int channel, void *address)
{
    assert(!dma_enabled, (void) 0);
    dma_address = (u16 *) address;
}

void set_dma_count(int channel, int count)
{
    assert(!dma_enabled, (void) 0);
    dma_count = 0;
    dma_residue = count;
}

void ccd_sim_initialise(void)
{
    int column, row;

    for (column = 0; column < CCD_READOUT_REGISTER_LENGTH; column++) {
	ccd_readout[column] = column;
    }

    for (row = 0; row < CCD_PHYSICAL_YDIM; row++) {
	for (column = 0; column < CCD_PHYSICAL_XDIM; column++) {
	    ccd_chip[row][column] = column + row;
	}
    }

    ccd_fifo_in = 0;
    ccd_fifo_out = 0;
    simd.interface_status = 0x80;
    simi.interface_status = 1;
}

void ccd_isr(int irq, void *dev_id, struct pt_regs *regs);

void ccd_test_isr(void)
{
    unsigned long flags;

    save_flags(flags);
    cli();
    ccd_isr(0, NULL, NULL);
    restore_flags(flags);
}

void ccd_sim_horizontal(u16 speed, u16 digitise,
			u16 continuous, u16 count, u16 binning)
{

/*
  Perform "count" horizontal shifts, using a binning factor of
  "binning".  A binning of 0 implies no binning (as per the Wright
  Instruments hardware definition).  The readout rate is high if
  "speed" is non-zero. The data are digitised if "digitise" is
  non-zero. The pixel reads/skips are continuous if "continuous" is
  true.

  Michael Ashley / UNSW / 09-May-2001 
*/

    int i, j, pos, pixels_to_zero, pixels_to_move;
    int num_zeroes = 0;

    assert(!(digitise && continuous), (void) 0);

    if (assertion_has_failed) {
	return;
    }

    /* Make sure we aren't already performing a vertical or horizontal
       operation. */

    assert((simd.status & 0x0180) == 0, (void) 0);

    simd.status |= 0x0080;	// Set the "performing horizontal operation" bit.

    /* Make "binning" more logical, ie, a value of 1 implies 1 pixel per bin. */

    binning++;

    /* Add up all the pixels being read out. */

    if (digitise) {
	// DANGER dma is not being cleared when using port i/O
	//        assert (simi.dma, (void) 0);
	for (i = 0; i < count; i++) {
	    ccd_output = 0;

	    for (j = 0; j < binning; j++) {
		pos = i * binning + j;
		if (pos < CCD_READOUT_REGISTER_LENGTH) {
		    ccd_output += ccd_readout[pos];
		} else {
		    assert(0, (void) 0);
		}
	    }

	    if (simd.dma) {	/* Transfer data using DMA */
		assert(dma_enabled && dma_address != NULL
		       && dma_residue != 0, (void) 0);
		*(dma_address++) = ccd_output;
		if (!ccd_output)
		    num_zeroes++;
		dma_count += 2;
		dma_residue -= 2;
		if (dma_residue == 0) {
		    assert(simi.cam_int && simd.cam_int, (void) 0);
		    ccd_test_isr();
		}
	    } else {		/* Transfer data using port I/O */
		ccd_fifo[ccd_fifo_in++] = ccd_output;
		if (ccd_fifo_in == CCD_FIFO_SIZE) {
		    ccd_fifo_in = 0;
		}
		assert(ccd_fifo_in != ccd_fifo_out, (void) 0);
	    }
	}
    }

    if (!continuous) {

	/* Now shift the readout register. */

	pixels_to_zero = MIN(count * binning, CCD_READOUT_REGISTER_LENGTH);
	pixels_to_move = CCD_READOUT_REGISTER_LENGTH - pixels_to_zero;

	memmove(ccd_readout, &(ccd_readout[pixels_to_zero]),
		(sizeof ccd_readout[0]) * pixels_to_move);
	memset(&(ccd_readout[pixels_to_move]), 0,
	       (sizeof ccd_readout[0]) * pixels_to_zero);

    } else {
	memset(ccd_readout, 0, sizeof ccd_readout);
    }

    //  printk ("horizontal %d %d %d %d %d %d\n", speed, digitise, continuous, count, binning, num_zeroes);
    simd.status &= 0xff7f;	// Clear "performing horizontal operation" bit
}

void ccd_sim_vertical(int image, int store, u16 count, int direction)
{
/*
  Simulate a vertical shift of the CCD. This is done by moving data around
  in a 2MB internal array. The readout register is handled too.

  "image"     is true if we are clocking the "image" section of 
                 the CCD (the "top" half).
  "store"     is true if we are clocking the "store" section of 
                 the CCD (the "bottom" half).
  "count"     is the number of vertical operations.
  "direction" is 0 to clock downwards (towards the readout register), and
                 1 to clock upwards.

  NOTE: currently, only downward movements of both image and store are
        implemented.

  Michael Ashley / UNSW / 14-May-2001
*/

    int bytes_to_copy, column, row;

    /* Make sure we aren't already performing a vertical or horizontal
       operation. */

    assert((simd.status & 0x0180) == 0, (void) 0);

    if (assertion_has_failed) {
	return;
    }

    simd.status |= 0x0100;	// Set the "vertical operation in progress" bit.

    if (direction) {		/* UP */
	if (image && store) {
	} else {
	}
	assert(0, (void) 0);
    } else {			/* DOWN */
	if (image && store) {
	    bytes_to_copy = (sizeof ccd_chip[0][0]) * CCD_PHYSICAL_XDIM
		* (CCD_PHYSICAL_YDIM - count);

	    // If we are reading out more than the physical number of rows in
	    // the CCD, then we can simply saturate the readout register and
	    // clear the chip.

	    if (bytes_to_copy < 0) {
		bytes_to_copy = 0;
		for (column = CCD_READOUT_REGISTER_OFFSET;
		     column < CCD_READOUT_REGISTER_OFFSET +
		     CCD_PHYSICAL_XDIM; column++) {
		    ccd_readout[column] = CCD_SATURATION;
		}
		memset(ccd_chip, 0, (sizeof ccd_chip));

		// Else we are reading out part of the chip. Add the relevant rows
		// to the readout register, and then move the CCD data and zero
		// the newly opened rows.

	    } else {
		for (row = 0; row < count; row++) {
		    for (column = 0; column < CCD_PHYSICAL_XDIM; column++) {
			ccd_readout[column +
				    CCD_READOUT_REGISTER_OFFSET] +=
			    ccd_chip[row][column];
		    }
		}
		if (count != 0) {
		    memmove(ccd_chip, &(ccd_chip[count][0]),
			    bytes_to_copy);
		    memset(&(ccd_chip[CCD_PHYSICAL_YDIM - count][0]), 0,
			   (sizeof ccd_chip) - bytes_to_copy);
		}
	    }
	} else {
	    assert(0, (void) 0);
	}
    }
    simd.status &= 0xfeff;	// Clear the "vertical operation in progress" bit.
}

unsigned ccd_sim(unsigned data, unsigned port, int direction)
{

    int vertical_mode, vertical_delay;
    int horizontal_mode;
    u16 vertical_count, horizontal_count, horizontal_binning;
    unsigned result;

    if (assertion_has_failed) {
	return 0;
    }

    result = 0;

    switch (direction) {
    case 0:			/* write */
	switch (port) {
	case CCD_HARDWARE_PORT:	/* send data word to camera */
	    switch ((data & 0xff00) >> 8) {
	    case 0:		/* select register/page */
		switch ((data & 0xffc0) >> 7) {
		case 0:	/* select register */
		    assert((data & 0x20) == 0, (void) 0);
		    simd.reg = data & 0x1f;
		    simi.reg = 1;
		    break;
		case 1:	/* select page */
		    simd.page = data & 0x3f;
		    simi.page = 1;
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 1:		/* write register */
		assert(simi.page && simi.reg, (void) 0);
		assert((simd.page < 64) && (simd.reg < 32), (void) 0);
		simd.par[simd.page][simd.reg] = data & 0x00ff;
		simi.par[simd.page][simd.reg] = 1;
		break;
	    case 2:		/* vertical trigger */
		switch (data & 0xff) {
		case 0:	/* abort vertical transfers */
		    break;
		case 1:	/* start vertical transfer */
		    assert(simi.page, (void) 0);

		    assert(simi.par[simd.page][0], (void) 0);
		    assert(simi.par[simd.page][2], (void) 0);
		    assert(simi.par[simd.page][3], (void) 0);
		    assert(simi.par[simd.page][4], (void) 0);
		    assert(simi.par[simd.page][5], (void) 0);

		    vertical_mode = simd.par[simd.page][0];
		    vertical_count =
			simd.par[simd.page][2] +
			256 * simd.par[simd.page][3];
		    vertical_delay =
			simd.par[simd.page][4] +
			256 * simd.par[simd.page][5];
		    //      printk ("vertical\n");
		    ccd_sim_vertical(vertical_mode & 0x01,
				     vertical_mode & 0x02,
				     vertical_count, vertical_mode & 0x04);
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 3:		/* horizontal trigger */
		switch (data & 0xff) {
		case 0:	/* abort horizontal transfers */
		    break;
		case 1:	/* start horizontal transfer */
		    assert(simi.page, (void) 0);
		    assert(simi.par[simd.page][16] &
			   simi.par[simd.page][18] &
			   simi.par[simd.page][19] &
			   simi.par[simd.page][20] &
			   simi.par[simd.page][21], (void) 0);
		    horizontal_mode = simd.par[simd.page][16];
		    horizontal_count =
			simd.par[simd.page][18] +
			256 * simd.par[simd.page][19];
		    horizontal_binning =
			simd.par[simd.page][20] +
			256 * simd.par[simd.page][21];
		    ccd_sim_horizontal(horizontal_mode & 0x04,
				       horizontal_mode & 0x02,
				       horizontal_mode & 0x01,
				       horizontal_count,
				       horizontal_binning);
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 4:		/* analog gain control / data shift */
		switch ((data & 0x80) >> 7) {
		case 0:	/* analog gain */
		    switch (data & 0x7f) {
		    case 0:
			simd.gain = 1;
			simi.gain = 1;
			break;
		    case 1:
			simd.gain = 4;
			simi.gain = 1;
			break;
		    default:
			assert(0, (void) 0);
		    }
		    break;
		case 1:	/* data shift */
		    switch ((data & 0x40) >> 6) {
		    case 0:	/* right shift */
			assert(!(data & 0x30), (void) 0);
			simd.shift = data & 0x0f;
			simi.shift = 1;
		    case 1:	/* left shift */
			assert(!(data & 0x30), (void) 0);
			simd.shift = data & 0x0f;
			simi.shift = 1;
			break;
		    default:
			assert(0, (void) 0);
		    }
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 5:
	    case 6:
	    case 7:
		assert(0, (void) 0);
		break;
	    case 8:		/* shutter control */
		switch (data & 0x00ff) {
		case 0:
		    simd.shutter = 0;
		    simi.shutter = 1;
		    break;
		case 1:
		    simd.shutter = 1;
		    simi.shutter = 1;
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 9:
		assert(0, (void) 0);
		break;
	    case 10:		/* cooler/fan control */
		assert(!(data & 0x00fc), (void) 0);
		switch (data & 0x0001) {
		case 0:
		    simd.cooler &= 0xfffe;
		    simi.cooler = 1;
		    break;
		case 1:
		    simd.cooler |= 0x0001;
		    simi.cooler = 1;
		    break;
		default:
		    assert(0, (void) 0);
		}
		switch ((data & 0x0002) > 1) {
		case 0:
		    simd.cooler &= 0xfffd;
		    simi.cooler = 1;
		    break;
		case 1:
		    simd.cooler |= 0x0002;
		    simi.cooler = 1;
		    break;
		default:
		    assert(0, (void) 0);
		}
		break;
	    case 11:		/* status return */
		ccd_fifo[ccd_fifo_in++] = simd.status;
		if (ccd_fifo_in == CCD_FIFO_SIZE) {
		    ccd_fifo_in = 0;
		}
		assert(ccd_fifo_in != ccd_fifo_out, (void) 0);
		break;
	    case 12:		/* status reset */
		simd.status = 0x8000;
		simi.status = 1;
		break;
	    default:
		assert(0, (void) 0);
		break;
	    }
	case CCD_HARDWARE_PORT + 2:	/* reset interface status to no error state */
	    simd.interface_status = 0x80;
	    simi.interface_status = 1;
	    break;
	case CCD_HARDWARE_PORT + 4:	/* enable/disable DMA */
	    switch (data) {
	    case 0:
		simd.dma = 0;
		simi.dma = 1;
		break;
	    case 1:
		simd.dma = 1;
		simi.dma = 1;
		break;
	    default:
		assert(0, (void) 0);
	    }
	    break;
	case CCD_HARDWARE_PORT + 6:	/* enable/disable camera interrupts */
	    switch (data) {
	    case 0:
		simd.cam_int = 0;
		simi.cam_int = 1;
		break;
	    case 1:
		simd.cam_int = 1;
		simi.cam_int = 1;
		break;
	    default:
		assert(0, (void) 0);
	    }
	    break;
	case CCD_HARDWARE_PORT + 8:	/* cancel camera interrupt request */
	    break;
	case CCD_HARDWARE_PORT + 10:	/* send data word to parallel interface */
	    simd.pll = data;
	    simi.pll = 1;
	    break;
	case CCD_HARDWARE_PORT + 12:	/* enable/disable parallel interface interrupts */
	    switch (data) {
	    case 0:
		simd.pll_int = 0;
		simi.pll_int = 1;
		break;
	    case 1:
		simd.pll_int = 1;
		simi.pll_int = 1;
		break;
	    default:
		assert(0, (void) 0);
	    }
	    break;
	case CCD_HARDWARE_PORT + 14:	/* cancel parallel interface interrupt request */
	    break;
	case CCD_HARDWARE_PORT + 16:	/* offset to add to data */
	    simd.offset = data;
	    simi.offset = 1;
	    break;
	default:
	    assert(0, (void) 0);
	}
	break;


    case 1:			/* read */
	switch (port) {
	case CCD_HARDWARE_PORT:	/* read next word from FIFO */
	    if (ccd_fifo_in != ccd_fifo_out) {
		result = ccd_fifo[ccd_fifo_out++];
		if (ccd_fifo_out == CCD_FIFO_SIZE) {
		    ccd_fifo_out = 0;
		}
	    } else {
                result = 0; /* DANGER: I'm not sure what the hardware
                               does when there isn't a word in the FIFO */
	    }
	    break;
	case CCD_HARDWARE_PORT + 2:	/* read interface status word */
	    if (ccd_fifo_in != ccd_fifo_out) {
		simd.interface_status |= 0x20;	// Set "ready to send" bit
	    } else {
		simd.interface_status &= 0xdf;	// Clear "ready to send" bit
	    }
	    result = simd.interface_status;
	    break;
	case CCD_HARDWARE_PORT + 10:	/* read data from parallel interface */
	    result = simd.pll;
	    break;
	default:
	    assert(0, (void) 0);
	}
	break;
    default:
	assert(0, (void) 0);
    }

    return result;
}

unsigned my_inw_p(unsigned port)
{
    in_this_time = jiffies;
    in_delta_time = in_this_time - in_last_time;
    in_last_time = in_this_time;
    return ccd_sim(0, port, 1);
}

void my_outw_p(unsigned data, unsigned port)
{
    out_this_time = jiffies;
    out_delta_time = out_this_time - out_last_time;
    out_last_time = out_this_time;
    ccd_sim(data, port, 0);
}


#endif				/* SIMULATOR */

// LOWEST-LEVEL ROUTINES FOR ACCESSING THE CCD ------------------------------

int ccd_send_command(unsigned int command)
{

/*
   Checks that the CCD is ready to receive a command, and then
   sends "command" to the CCD data port. If the CCD is not ready,
   up to CCD_RECEIVE_TIMEOUT attempts will be made before
   giving up.

   When the command is sent, the global variable "last_command"
   is set to "command", and "last_command_number" is incremented.

   NOTE: this routine is not reentrant, and should not be
   interrupted by a routine that accesses the CCD. The 
   reason for this is that the CCD may become busy between
   CCD_IS_READY_TO_RECEIVE becoming true, and the command
   being sent.

   To assist with ensuring the above, this routine MUST ONLY BE
   CALLED from within "ccd_process_command ()".

   Returns:

   0             success, the command was sent
   -EWOULDBLOCK  failure, the CCD was not ready to receive the command

   Michael Ashley / UNSW / 08-Aug-1999
*/

    int i;

    for (i = 0; i < CCD_RECEIVE_TIMEOUT; i++) {
	if (CCD_IS_READY_TO_RECEIVE) {
#ifndef NDEBUG
	    // M0001 - Save only 'real' commands, because status return
	    // is done after a lot of commands,
	    // so this could 'hide' the last real one.
	    if (command != COMMAND_STATUS_RETURN)
		last_command = command;
	    last_command_number++;
#endif
	    outw_p(command + ccd_number, CCD_DATA_PORT);
	    //       printk ("%x ", command);
	    break;
	}
    }

    if (i == CCD_RECEIVE_TIMEOUT) {
	return -EWOULDBLOCK;
    }
    return 0;
}

int ccd_receive_word(u16 * inword)
{

/*
   Attempts to read a word (data or status) from the CCD.

   NOTE: this routine is non-reentrant, and should not be interrupted
   by routines that communicate with the CCD. It should only
   be called from within "ccd_process_command ()".

   Returns:

   0             success, the word was read (and is available in "inword")
   -EWOULDBLOCK  failure, a word was not available, or DMA was in
                 progress

   Michael Ashley / UNSW / 08-Aug-1999
*/


    int i;

    if (!ccd_mode.dma_in_progress) {
	for (i = 0; i < CCD_SEND_TIMEOUT; i++) {
	    if (CCD_IS_SENDING) {
		*inword = inw(CCD_DATA_PORT);
		// NOTE: we use inw() rather than
		//   inw_p(), since the latter takes 
		//   too long and does not appear to be 
		//   necessary.

		return 0;
	    }
	}
    }
    *inword = 0xFFFF;
    // So as not to lull the user into a 
    //   false sense of security, if the return 
    //   value is not checked.

    return -EWOULDBLOCK;
}

int ccd_read_status(void)
{

/*
   Reads and returns the CCD status word. If successful, the
   following global variables are set:

   ccd_mode.status <-- the CCD status
   ccd_init.status <-- true if ccd_mode.status is valid
   ccd_mode.time   <-- the UNIX time at which the status was read

   NOTE: this routine is non-reentrant, and should not be interrupted
   by routines that communicate with the CCD. It should only
   be called from within "ccd_process_command ()".

   Returns:

   >= 0          success, the CCD status is returned
   -EWOULDBLOCK  failure, the CCD was not ready

   Michael Ashley / UNSW / 08-Aug-1999
*/

    int ret;

    ccd_init.status = 0;	// M003 Begin with this false.
    if (0 == (ret = ccd_send_command(COMMAND_STATUS_RETURN))) {
	do_gettimeofday(&(ccd_mode.time));
	ret = ccd_receive_word(&(ccd_mode.status));
	ccd_init.status = !ret;
    }
    return ret;
}


// ROUTINES FOR PROCESSING THE DMA BUFFER

int dma_allocate_buffer(unsigned int min_bytes)
{

/*
   If min_bytes != 0, attempts to allocate a memory block suitable for
   DMA that is at least "min_bytes" in length. If successful, various
   global pointers will be set to point to the block.

   If min_bytes == 0, initialise global pointers to a DMA area that was
   previously allocated.

   Explanation of the DMA buffer area:

   All ADC data from the camera, whether obtained using DMA or port
   I/O, is placed in a ring-buffer area beginning at the location
   "dma_begin", and ending one byte before "dma_end". The buffer is
   obtained using the operating system call "__get_dma_pages()",
   guaranteeing that the memory is suitable for DMA (e.g., it is
   contiguous).

   The ring-buffer is complicated by the fact that DMA can't wrap,
   i.e., if you want to store more data than available between the
   current location and the end of the buffer, you must restart at the
   beginning of the buffer again. "dma_last_written_pointer" is used
   to keep track of the location within the buffer at which the wrap
   last occurred.

   The buffer is allocated during "init_module()" and deallocated in
   "cleanup_module()".

   The size of the buffer is governed by the enum 
   MAX_WORDS_PER_DMA_TRANSFER. See "ccd_find_hardware()" for details.

   Here are the various pointers used to maintain the integrity of the
   buffer.

   dma_begin                 the first byte of the buffer
   dma_read_pointer          the point at which a byte will next be read
                               (from the buffer to the user program)
   dma_next_read_pointer     the location to which dma_read_pointer should
                               be set to after the next read operation
   dma_write_pointer         the point at which the next byte from the
                               camera will be stored; for DMA it is location
                               at which the DMA transfer will start
   dma_next_write_pointer    one byte beyond the end of the data being 
                               written
   dma_last_written_pointer  one byte beyond the point at which data has
                               been written to in the buffer
   dma_end                   one byte beyond the end of the buffer

   Returns:

   0        success
   -ENOMEM  failure, the block could not be allocated

   Michael Ashley / UNSW / 11-Aug-1999 */

    int order, pages_required, slot;

    if (min_bytes != 0) {
	pages_required = 1 + min_bytes / PAGE_SIZE;
	order = -1;
	while (pages_required > 0) {
	    pages_required /= 2;
	    order++;
	}
	
	for (slot = 0; slot < 2; slot++) {

	    // Attempt to allocate the memory, exit if unsuccessful.

	    dma_begin[slot] = __get_dma_pages(GFP_KERNEL, order);

	    if (!dma_begin[slot]) {
		return -ENOMEM;
	    }

	    dma_end[slot] = dma_begin[slot] + (PAGE_SIZE << order);
	}
    } 

    // Initialise a bunch of global variables that point to important
    // locations within the DMA buffer.

    for (slot = 0; slot < 2; slot++) {
	dma_read_pointer[slot]         = dma_begin[slot];
	dma_next_read_pointer[slot]    = dma_begin[slot];
	dma_write_pointer[slot]        = dma_begin[slot];
	dma_next_write_pointer[slot]   = dma_begin[slot];
	dma_last_written_pointer[slot] = dma_begin[slot];
    }

    return 0;
}

int dma_prepare_to_write(int slot, unsigned int byte_count)
{

/*
   Sets up the various DMA pointers to write "byte_count" bytes
   from the CCD camera into the buffer.

   Returns:

   0             success
   -EINVAL       should only occur due to a programming error
   -EWOULDBLOCK  failure, there is no room currently available

   Michael Ashley / UNSW / 11-Aug-1999
*/

    unsigned long dma_1, dma_2;

    assert(slot == 0 || slot == 1, return -EINVAL);
    assert(!ccd_mode.dma_in_progress, return -EINVAL);
    assert(byte_count <= (dma_begin[slot] - dma_end[slot]),
	   return -EINVAL);

/* 
   We calculate the starting (dma_1) and ending (dma_2 - 1)
   locations within the DMA buffer that will be required to store
   "byte_count" bytes.  

   "dma_next_write_pointer" is the nominal starting location for this
   DMA operation. We check to ensure that there is room to write
   "byte_count" bytes before running into the end of the buffer. If
   not, we examine the possibility of starting at the beginning of the
   buffer.
 */

    dma_1 = dma_next_write_pointer[slot];
    if (dma_1 + byte_count > dma_end[slot]) {

	// Hm... there isn't room. We can only wrap if we will not overwrite
	// data that has not been read.

	if ((dma_write_pointer[slot] < dma_read_pointer[slot]) ||
	    (dma_read_pointer[slot] == dma_begin[slot])) {
	    ccd_mode.dma_overwrite++;
#if DEBUG_VERBOSE
	    printk("overwrite0: %ld %ld %ld %ld %ld\n",
		   dma_read_pointer[slot] - dma_begin[slot],
		   dma_write_pointer[slot] - dma_begin[slot],
		   dma_next_write_pointer[slot] - dma_begin[slot],
		   dma_end[slot] - dma_begin[slot],
		   dma_last_written_pointer[slot] - dma_begin[slot]);
#endif
	    return -EWOULDBLOCK;
	}
	dma_1 = dma_begin[slot];
    }
    dma_2 = dma_1 + byte_count;

    /*
       Now we examine whether this range of addresses encompasses the
       dma_read_pointer. If it does, we would overwrite data which has
       not yet been read! So we exit with -EWOULDBLOCK to let
       the calling program know that DMA is not a good idea at this
       point in time.
     */

    if ((dma_read_pointer[slot] > dma_1) &&
	(dma_read_pointer[slot] <= dma_2)) {
	ccd_mode.dma_overwrite++;
#if DEBUG_VERBOSE
	printk("overwrite1: %ld %ld %ld %ld %ld\n",
	       dma_read_pointer[slot] - dma_begin[slot],
	       dma_write_pointer[slot] - dma_begin[slot],
	       dma_next_write_pointer[slot] - dma_begin[slot],
	       dma_end[slot] - dma_begin[slot],
	       dma_last_written_pointer[slot] - dma_begin[slot]);
#endif
	return -EWOULDBLOCK;
    }
    if (dma_write_pointer[slot] > dma_1) {
	dma_last_written_pointer[slot] = dma_next_write_pointer[slot];
    }
    dma_write_pointer[slot] = dma_1;
    dma_next_write_pointer[slot] = dma_2;
#if DEBUG_VERBOSE
    printk("DMA OK: %ld %ld %ld %ld %ld\n",
	   dma_read_pointer[slot] - dma_begin[slot],
	   dma_write_pointer[slot] - dma_begin[slot],
	   dma_next_write_pointer[slot] - dma_begin[slot],
	   dma_end[slot] - dma_begin[slot],
	   dma_last_written_pointer[slot] - dma_begin[slot]);
#endif
    return 0;
}


void dma_prepare_to_read(int slot, size_t * byte_count)
{

/*
   Examines a request to read "byte_count" bytes from the DMA buffer.
   If fewer than this many bytes are available, "byte_count" is
   overwritten with the number available.

   After calling this routine, it is always possible to read
   "byte_count" bytes starting at "dma_read_pointer". 

   "dma_next_read_pointer" will contain the value to which 
   "dma_read_pointer" should be set after successfully reading the data.

   Michael Ashley / UNSW / 13-Aug-1999
 */

    unsigned long wp;		// write pointer

    size_t bc;			// byte count

    assert(slot == 0 || slot == 1, *byte_count = 0);

    // Grab a copy of the write pointer, to guard against changes due
    // to interrupts.

    wp = dma_write_pointer[slot];

    // If the write and read pointers are identical, there is no data
    // available.

    while (1) {
	if (wp == dma_read_pointer[slot]) {
	    *byte_count = 0;
	    return;

	    // If the write pointer is greater than the read pointer,
	    // we have the simple case where DMA has not wrapped to
	    // the beginning of the buffer.

	} else if (wp > dma_read_pointer[slot]) {
	    bc = wp - dma_read_pointer[slot];
	    if (bc < *byte_count) {
		*byte_count = bc;
	    }
	    dma_next_read_pointer[slot] =
		dma_read_pointer[slot] + *byte_count;
	    return;

	    // Else the write pointer must be less that the read pointer,
	    // in which case the DMA buffer has wrapped, and
	    // "dma_last_written_pointer" points to one byte beyond the
	    // last written location in the buffer.

	} else {
	    if (dma_last_written_pointer[slot] > dma_read_pointer[slot]) {
		bc = dma_last_written_pointer[slot] -
		    dma_read_pointer[slot];

		// If we can read all the way to the end, the next read
		// pointer will be at the beginning of the buffer.

		if (bc <= *byte_count) {
		    dma_next_read_pointer[slot] = dma_begin[slot];
		    *byte_count = bc;
		} else {
		    dma_next_read_pointer[slot] = dma_read_pointer[slot] +
			*byte_count;
		}
		return;

		// Else we have read right to the end of the written region
		// of the buffer, so we can restart from the beginning.

	    } else {
		assert(dma_last_written_pointer[slot] ==
		       dma_read_pointer[slot], (void) 0);
		dma_read_pointer[slot] = dma_begin[slot];
		dma_last_written_pointer[slot] = wp;	// EXTREME DANGER (new addition)
	    }
	}
    }
}

int ccd_dma_prepare(unsigned int byte_count)
{

/*
   Sets up the CCD for a DMA operation of "byte_count" bytes.
   The end of the operation will be signalled by an interrupt
   from the CCD.

   Returns:

   0             success
   -EINVAL       should only occur due to a programming error
   -EWOULDBLOCK  failure, the CCD was not ready

   Michael Ashley / UNSW / 11-Aug-1999 
 */

    unsigned long flags;
    int ret;

    // If the user has requested that the output file should be closed, we
    // don't allow any new DMA activity until the file is actually
    // closed by a call to "ccd_read".

    if (ccd_mode.closing_file) {
	return -EWOULDBLOCK;
    }
    RETURN_ON_FAILURE(dma_prepare_to_write(ccd_mode.dma_slot, byte_count));

    ccd_mode.dma_in_progress = 1;

    // Set up the PC's DMA interface. We do this with interrupts disabled
    // so that nothing interferes with programming the DMA channel.

    save_flags(flags);
    cli();

    disable_dma(dma);
    clear_dma_ff(dma);
    set_dma_mode(dma, DMA_MODE_READ);
    set_dma_addr(dma,
		 virt_to_bus((void *)
			     dma_write_pointer[ccd_mode.dma_slot]));
    set_dma_count(dma, byte_count);
    enable_dma(dma);

    // Set up the CCD camera's DMA interace.

    outw_p(0x0000, CCD_CANCEL_CAM_INT_PORT);
    outw_p(0x0001, CCD_CAM_INT_PORT);
    outw_p(0x0001, CCD_DMA_PORT);

#ifndef NDEBUG
    dma_starting_time = jiffies;
#endif

    // Set some global variables.

    ccd_mode.dma = 0x0001;
    ccd_init.dma = 1;

    restore_flags(flags);
    return 0;
}

void ccd_dma_cancel(void)
{
    outw_p(0x0000, CCD_CAM_INT_PORT);
    outw_p(0x0000, CCD_CANCEL_CAM_INT_PORT);
    outw_p(0x0000, CCD_DMA_PORT);
}

void ccd_isr(int irq, void *dev_id, struct pt_regs *regs)
{

    // This is the interrupt service routine to handle interrupts 
    // generated by the CCD.

    // Schedule a bottom half to perform the more time-consuming tasks.
    // Note that we do not use the "_irq_off" variant of "queue_task", since
    // it disappeared in the 2.2 kernel.

    int i, word_count;
#ifndef NDEBUG
    int j;
#endif
    u16 data, *data_pointer;

    // Disable camera interrupts, cancel any outstanding interrupts,
    // disable camera DMA.

    ccd_dma_cancel();

    //  Reset global variables.

    ccd_mode.dma = 0x0000;
    ccd_mode.dma_in_progress = 0;
    ccd_mode.words_expected = 0;
    ccd_mode.readout_in_progress = 0;

#ifndef NDEBUG

    dma_delta_time = jiffies - dma_starting_time;
    for (i = 0; i < 10; i++) {
	if (dma_delta_time > dma_times[i]) {
	    for (j = 9; j >= i; j--) {
		dma_times[j + 1] = dma_times[j];
	    }
	    dma_times[i] = dma_delta_time;
	    break;
	}
    }
#endif

    if (throw_away_dma_data) {
	throw_away_dma_data = 0;
	return;
    }

    // We now subtract 32768 from all the data points and swap the lo and hi
    // bytes so that we conform with the FITS standard.

    word_count = (dma_next_write_pointer[ccd_mode.dma_slot] -
		  dma_write_pointer[ccd_mode.dma_slot]) / 2;
    data_pointer = (u16 *) dma_write_pointer[ccd_mode.dma_slot];

    for (i = 0; i < word_count; i++) {
	data = *data_pointer - 32768;
	*(data_pointer++) = htons(data);	// M0001
    }

    // Update the write pointer.

    dma_write_pointer[ccd_mode.dma_slot] =
	dma_next_write_pointer[ccd_mode.dma_slot];

    // M0001 - If cleanup_module() wants us to die do not reinstall ISR.

    if (!module_shutting_down) {
	ccd_queue_service_task(1);
    }
}

struct timer_list wait_timer, cleanup_timer;

int ccd_process_command(struct ccd_command_struct *p)
{
/*
   Given a pointer "p" to a "ccd_command_struct", this routine
   processes the command. 

   All commands sent to the hardware should originate from this
   routine.  Note: this is violated only for ccd_find_hardware().

   This routine is guaranteed to return within a reasonable length of
   time (a millisecond or so?), having either completed the command,
   initiated a sequence of operations (e.g., DMA readout), or having
   been unable to complete the command. In the latter case, you can
   simply retry until success occurs.

   This routine is not reentrant. If it is used in an ISR,
   you must guarantee that the interrupt can not occur within this
   routine.

   Commands come in the following flavours:

   (1) Simple commands sent to the CCD data port, e.g., selecting a
       page of parameter memory.

   (2) Simple commands sent to other CCD ports, e.g., communicating
       with the parallel port.

   (3) More complex commands, that may involve a set of transactions
       with the CCD, e.g., waiting for vertical transfers to stop.

   (4) Highly complex things, e.g., reading out part of the CCD,
       whether by using DMA or port I/O; timed shutter control.

   Return values:

   0             success (i.e., the command was successfully sent to the CCD).
   -EINVAL       a programming error has occurred.
   -EWOULDBLOCK  the command was not able to be fully processed. In all 
                 cases it is OK to try again.

   The possible commands are:

   COMMAND_SELECT_REGISTER
   COMMAND_WRITE_REGISTER
   COMMAND_HORIZONTAL_TRIGGER
   COMMAND_VERTICAL_TRIGGER
   COMMAND_SHUTTER
   COMMAND_COOLER
   COMMAND_ANALOG_CONTROL
   COMMAND_STATUS_RETURN
   COMMAND_STATUS_RESET

   CCD_STATUS_PORT
   CCD_DMA_PORT
   CCD_CAM_INT_PORT
   CCD_CANCEL_CAM_INT_PORT
   CCD_PLL_PORT
   CCD_PLL_INT_PORT
   CCD_CANCEL_PLL_INT_PORT
   CCD_OFFSET_PORT

   CCD_STOP_HORIZONTAL_TRANSFERS
   CCD_WAIT_FOR_HORIZONTAL_TRANSFERS
   CCD_STOP_VERTICAL_TRANSFERS
   CCD_WAIT_FOR_VERTICAL_TRANSFERS

   Careful testing is done to ensure that the command makes sense
   given the previous history of the CCD (e.g., you will be unable
   to write to a register until a register has been selected; if you
   try to do so, this module will write a message to the kernel log, and
   become disabled).

   Michael Ashley / UNSW / 02-Jan-1999 
*/

    int ret, i;

    if (assertion_has_failed)
	return -EINVAL;

    assert(ccd_setup_to_use_interrupts &&
	   !ccd_mode.dma_in_progress && servicing, return -EINVAL);

//  printk ("command 0x%x 0x%x 0x%x\n", p->port, p->delay, p->outword);

    switch (p->port * 2) {
    case CCD_DATA_PORT - io:
	switch (p->outword & COMMAND_MASK) {

	case COMMAND_SELECT_REGISTER:
	    if (p->outword & BIT_PAGE) {
		assert(!(p->outword & 0x0040), return -EINVAL);
	    } else {
		assert(!(p->outword & 0x00e0), return -EINVAL);
	    }
	    RETURN_ON_FAILURE(ccd_send_command(p->outword));
	    if (p->outword & BIT_PAGE) {
		ccd_mode.page = p->outword & 0x003f;
		ccd_init.page = 1;
	    } else {
		ccd_mode.reg = p->outword & 0x001f;
		ccd_init.reg = 1;
	    }
	    break;

	case COMMAND_WRITE_REGISTER:
	    assert(ccd_init.page && ccd_init.reg, return -EINVAL);

	    switch (ccd_mode.reg) {

	    case PARAMETER_VERTICAL_MODE:
	    case PARAMETER_HORIZONTAL_MODE:
		assert(!(p->outword & 0x00f0), return -EINVAL);
		break;

	    case PARAMETER_VERTICAL_REPEAT_COUNT_LO:
	    case PARAMETER_VERTICAL_REPEAT_COUNT_HI:
	    case PARAMETER_VERTICAL_DELAY_LO:
	    case PARAMETER_VERTICAL_DELAY_HI:
	    case PARAMETER_HORIZONTAL_REPEAT_COUNT_LO:
	    case PARAMETER_HORIZONTAL_REPEAT_COUNT_HI:
	    case PARAMETER_HORIZONTAL_BINNING_LO:
	    case PARAMETER_HORIZONTAL_BINNING_HI:
		break;

	    default:
		assert(!(ccd_mode.reg & 0xffe0)
		       && !p->outword, return -EINVAL);
	    }

	    RETURN_ON_FAILURE(ccd_send_command(p->outword));

	    assert(VALID__PAGE && VALID_REG, return -EINVAL);

	    CURRENT_PAGE_PAR[ccd_mode.reg] = (unsigned char)
		(p->outword & 0x00ff);
	    CURRENT_PAGE_INIT[ccd_mode.reg] = 1;
	    break;

	case COMMAND_HORIZONTAL_TRIGGER:
	    assert((((p->outword & 0x00ff) == 0x0001) && VALID__PAGE &&
		    READY_FOR_HORIZONTAL_TRIGGER) ||
		   ((p->outword & 0x00ff) == 0x0000), return -EINVAL);
	    if (DIGITIZING) {
		Expdata.speed =
		    (CURRENT_PAGE_PAR[PARAMETER_HORIZONTAL_MODE]
		     & 0x0004) >> 2;
		ccd_mode.words_expected =
		    CURRENT_PAGE_PAR[PARAMETER_HORIZONTAL_REPEAT_COUNT_LO]
		    +
		    CURRENT_PAGE_PAR[PARAMETER_HORIZONTAL_REPEAT_COUNT_HI]
		    * 256;

		if (ccd_mode.using_dma) {
		    assert(ccd_mode.words_expected <=
			   MAX_WORDS_PER_DMA_TRANSFER, return -EINVAL);
		    if (0 !=
			(ret =
			 ccd_dma_prepare(ccd_mode.words_expected * 2))) {
			ccd_mode.words_expected = 0;
			return ret;
		    }
		}
		ccd_mode.readout_in_progress = 1;
	    }
	    if (0 != (ret = ccd_send_command(p->outword))) {
		if (ccd_mode.dma_in_progress) {
		    ccd_dma_cancel();
		}
		ccd_mode.words_expected = 0;
		ccd_mode.readout_in_progress = 0;
		return ret;
	    }
	    break;

	case COMMAND_VERTICAL_TRIGGER:
	    assert((((p->outword & 0x00ff) == 0x0001) && VALID__PAGE &&
		    READY_FOR_VERTICAL_TRIGGER) ||
		   ((p->outword & 0x00ff) == 0x0000), return -EINVAL);
	    RETURN_ON_FAILURE(ccd_send_command(p->outword));
	    break;

	case COMMAND_SHUTTER:
	    assert(!(p->outword & 0x00fe), return -EINVAL);

	    if (p->outword & 0x0001) { // open shutter
		RETURN_ON_FAILURE(ccd_send_command(p->outword));
		do_gettimeofday(&(Expdata.exp_start_time));
		printk("ccd: opening shutter at %ld.%06ld\n", 
		       Expdata.exp_start_time.tv_sec, 
		       Expdata.exp_start_time.tv_usec);
#ifdef DEBUG_VERBOSE
		{struct timeval now;
		do_gettimeofday(&now);
		printk("ccd: now0 %ld.%06ld\n",
		       now.tv_sec, now.tv_usec);
		} 
#endif
		ccd_mode.shutter = 1;

	    } else { // close shutter
		RETURN_ON_FAILURE(ccd_send_command(p->outword));
		do_gettimeofday(&(Expdata.exp_end_time));
		printk("ccd: closing shutter at %ld.%06ld\n",
		       Expdata.exp_end_time.tv_sec, 
		       Expdata.exp_end_time.tv_usec);
#ifdef DEBUG_VERBOSE
		{struct timeval now;
		do_gettimeofday(&now);
		printk("ccd: now1 %ld.%06ld\n",
		       now.tv_sec, now.tv_usec);
		}
#endif
		ccd_mode.shutter = 0;
	    }
	    ccd_init.shutter = 1;
	    break;

	case COMMAND_COOLER:
	    assert(!(p->outword & 0x00fc), return -EINVAL);
	    RETURN_ON_FAILURE(ccd_send_command(p->outword));
	    ccd_mode.cooler = p->outword & 0x0003;
	    ccd_init.cooler = 1;	// M00001
	    break;

	case COMMAND_ANALOG_CONTROL:
	    if (p->outword & BIT_SHIFT) {
		assert(!(p->outword & 0x0030), return -EINVAL);
	    } else {
		assert(!(p->outword & 0x007e), return -EINVAL);
	    }
	    RETURN_ON_FAILURE(ccd_send_command(p->outword));
	    if (p->outword & BIT_SHIFT) {
		ccd_mode.shift = p->outword & 0x004f;
		ccd_init.shift = 1;
	    } else {
		ccd_mode.gain = p->outword & 0x0001;
		ccd_init.gain = 1;
	    }
	    break;

	case COMMAND_STATUS_RETURN:
	    assert(!(p->outword & 0x00ff), return -EINVAL);
	    RETURN_ON_FAILURE(ccd_read_status());
	    break;

	case COMMAND_STATUS_RESET:
	    assert(!(p->outword & 0x00ff), return -EINVAL);
	    RETURN_ON_FAILURE(ccd_send_command(p->outword));
	    break;

	default:
	    assert(0, return -EINVAL);
	}
	break;

    case CCD_STATUS_PORT - io:
	assert(p->outword == 0, return -EINVAL);
	outw_p(p->outword, p->port * 2 + io);
	break;

    case CCD_DMA_PORT - io:
	assert(!(p->outword & 0x00fe), return -EINVAL);
	outw_p(p->outword, p->port * 2 + io);
	ccd_mode.dma = p->outword;
	ccd_init.dma = 1;
	break;

    case CCD_CAM_INT_PORT - io:
	assert(!(p->outword & 0x00fe), return -EINVAL);
	outw_p(p->outword, p->port * 2 + io);
	ccd_mode.cam_int = p->outword;
	ccd_init.cam_int = 1;
	break;

    case CCD_CANCEL_CAM_INT_PORT - io:
	assert(p->outword == 0, return -EINVAL);
	outw_p(p->outword, p->port * 2 + io);
	break;

    case CCD_PLL_PORT - io:
	outw_p(p->outword, p->port * 2 + io);
	ccd_mode.pll = p->outword;
	ccd_init.pll = 1;
	break;

    case CCD_PLL_INT_PORT - io:
	outw_p(p->outword, p->port * 2 + io);
	ccd_mode.pll_int = p->outword;
	ccd_init.pll_int = 1;
	break;

    case CCD_CANCEL_PLL_INT_PORT - io:
	assert(p->outword == 0, return -EINVAL);
	outw_p(p->outword, p->port * 2 + io);
	break;

    case CCD_OFFSET_PORT - io:
	outw_p(p->outword, p->port * 2 + io);
	ccd_mode.offset = p->outword;
	ccd_init.offset = 1;
	break;

    default:
	assert(p->outword == 0, return -EINVAL);
	switch (p->port) {

	    // NOTE: we don't have a break after the next case, since we
	    // want control to continue.

	case CCD_STOP_HORIZONTAL_TRANSFERS:
	    RETURN_ON_FAILURE(ccd_send_command
			      (COMMAND_HORIZONTAL_TRIGGER));

	case CCD_WAIT_FOR_HORIZONTAL_TRANSFERS:
	    for (i = 0; i <= p->delay; i++) {
		RETURN_ON_FAILURE(ccd_read_status());
		if (ccd_init.status && !(ccd_mode.status & BIT_HORIZONTAL)) {
		    break;
		}
	    }
	    if (i > p->delay) {
		return -EWOULDBLOCK;
	    }
	    break;

	    // NOTE: we don't have a break after the next case, since we
	    // want control to continue.

	case CCD_STOP_VERTICAL_TRANSFERS:
	    RETURN_ON_FAILURE(ccd_send_command(COMMAND_VERTICAL_TRIGGER));

	case CCD_WAIT_FOR_VERTICAL_TRANSFERS:
	    for (i = 0; i <= p->delay; i++) {
		RETURN_ON_FAILURE(ccd_read_status());
		if (ccd_init.status && !(ccd_mode.status & BIT_VERTICAL)) {
		    break;
		}
	    }
	    if (i > p->delay) {
		return -EWOULDBLOCK;
	    }
	    break;

	default:
	    assert(0, return -EINVAL);
	}
    }
    return 0;
}

int ccd_extract_number(unsigned int *number)
{

/*
   Attempts to decode an integer number immediately after the current 
   position in the output ring buffer.

   A number is defined as being from 1 to 7 consecutive characters
   chosen from the set "0123456789". Any character outside this range
   will terminate the number. The number itself is returned in
   "number".  If the number has no characters in it, it is set to zero.

   Return values:

   > 0       success, the return value is the number of
             characters read (which is one more than the number of
             characters in the number)
   -EINVAL   the number exceeded 7 characters in length
   -ENODATA  the end of the ring buffer was reached before 
             the number terminated

   Examples:

   In the following table, the current position in the ring buffer is
   the letter "d" at the left.

   Ring buffer  Return        Number
   ---------------------------------

   dx              1               0
   d123x           4             123
   d           -ENODATA 
   d1234       -ENODATA      
   d12345678d  -EINVAL
   
   Michael Ashley / UNSW / 08-Aug-1999
*/

    int pos;
    int chars_consumed;
    unsigned char c;

    *number = 0;
    chars_consumed = 1;

    pos = (ccd_ring_o_tail + 1) % CCD_RING_O_BUFFER_LENGTH;

    while (pos != ccd_ring_o_head) {
	c = ccd_ring_o_buffer[pos];
	if (c < '0' || c > '9') {
	    return chars_consumed;
	}
	*number = *number * 10 + (c - '0');
	if (chars_consumed++ == 9) {
	    assert(0, (void) 0);
	    return -EINVAL;
	}
	pos = (pos + 1) % CCD_RING_O_BUFFER_LENGTH;
    }
    return -ENODATA;
}

#define is_digit(c) ((c) >= '0' && (c) <= '9')
//----------------------------------------------------------------------------
// Convert 'str' to integer.
//----------------------------------------------------------------------------
int atoi(char *str)
{
    int i, val = 0, sign = 1;

    for (i = 0; i < strlen(str); i++) {
	// Skip leading white space
	if ((str[i] == ' ' || str[i] == '\t') && val == 0)
	    continue;
	if (is_digit(str[i]))
	    val = val * 10 + str[i] - '0';
	else if (val == 0) {
	    // Leading sign or break out
	    if (str[i] == '+');
	    else if (str[i] == '-')
		sign = -1;
	    else
		break;
	} else
	    break;
    }
    return val * sign;
}

//----------------------------------------------------------------------------
// This function handles the different CCD type and FITS header initialization
// parameters. The command is introduced by the "I" command character and
// dispatched by ccd_service(). The format is as follows:
//  "I" parameter_selector  value
//      -------------------------
//      a  - altitude       string
//      bs - observer       string
//      bi - binning        int
//      c  - camera type    int
//      d  - DEC            string
//      em - email          string
//      ex - end x          int
//      ey - end y          int
//      f  - fits lines     string (max 6 lines of free stype FITS header stuff)
//      in - institute      string
//      im - image type     string (e.g.: object, bias, flat, dark)
//      j  - object         string
//      l  - latitude       string
//      n  - longitude      string
//      o  - observatory    string
//      p  - plate scale    string
//      ra - RA             string
//      re - reset all      string (contents irrelevant)
//      su - subset         string
//      sx - start x        int
//      sy - start y        int
//      t  - telescope      string
//      u  - instrument     string
//      w  - www address    string
//      x  - x pixel size   string
//      y  - y pixel size   string
//      z  - time zone      int
// Value MUST be enclosed in double quotes (") in order to easily find 
// their ends.
// Double quotes in value MUST be escaped with a '\'.
// M0002 - changed from single quotes to double quotes since single quotes
//         appear often in FITS headers, whereas double ones don't.
//----------------------------------------------------------------------------
#define NEXT_POS(p)     ((p + 1) % CCD_RING_O_BUFFER_LENGTH)
int ccd_parameters(int *nchars)
{
    int pos, sel, consumed, index, swallow, ok, act, prev;
    int mod = '\0';
    char cv[sizeof(Expdata.obs)], *p;

    *nchars = 1;
    pos = NEXT_POS(ccd_ring_o_tail);
    if (pos != ccd_ring_o_head) {
	index = 0;
	swallow = 0;
	sel = ccd_ring_o_buffer[pos];
	pos = NEXT_POS(pos);
	*nchars += 1;
	consumed = 2;
	if (sel == 'r' || sel == 'i' || sel == 'b' ||
	    sel == 's' || sel == 'e') {
	    // These commands are built up by two char selectors,
	    // collect it.
	    mod = ccd_ring_o_buffer[pos];
	    pos = NEXT_POS(pos);

	    // Dear Lazac: I am worried by this routine. Without looking at it in
	    // detail, it seems that you run the risk of overflowing the
	    // ccd_ring_o_buffer. For example, at this point you have just
	    // grabbed another character from the buffer without checking that
	    // the tail has wrapped around to the head. We need to be very
	    // careful about these sorts of bugs, since they are likely to cause
	    // intermittent crashes that are very hard to find.

	    *nchars += 1;
	    consumed++;
	}
	assert(ccd_ring_o_buffer[pos] == '\"', return -EINVAL);
	pos = NEXT_POS(pos);	// Skip leading quote
	consumed++;
	ok = 0;			// Assume NOT ok
	prev = 0;
	while (pos != ccd_ring_o_head) {	// Collect value
	    act = ccd_ring_o_buffer[pos];
	    pos = NEXT_POS(pos);
	    consumed++;
	    if (!swallow) {
		cv[index++] = act;
		if (act == '\"' && prev != '\\') {
		    // End of string
		    cv[index - 1] = '\0';
		    ok = 1;
		    break;
		}
		// Dear Lazac, here you were comparing "index == 41",
		// and then writing into cv[41], which is 1 greater than
		// the allowed size of cv. I haven't checked thoroughly
		// for similar bugs.

		// In fact, cv must be the same size as the Expdata
		// fields, since otherwise the trailing null byte on one
		// field will overwrite the beginning of the next one in
		// the case that cv fills up.

		// If at all possible, use sizeof() rather than put explicit
		// numbers in.

		if (index == sizeof(cv) - 1) {	// Swallow rest
		    cv[index] = '\0';
		    swallow = 1;
		}
	    } else if (act == '\"' && prev != '\\') {
		ok = 1;
		break;
	    }
	    prev = act;
	}
	*nchars = consumed;

	// Note, "strcpy" is a very dangerous routine, you need to
	// be absolutely sure that the string you are copying from is
	// null terminated at the right length. 

	assert(strlen(cv) < sizeof(Expdata.obs), return -EINVAL);

	if (ok) {
	    switch (sel) {
	    case 'a':
		strcpy(Expdata.alt, cv);
		break;
	    case 'b':
		if (mod == 's')
		    strcpy(Expdata.observer, cv);
		else if (mod == 'i')
		    Expdata.xb = atoi(cv);
		break;
	    case 'c':
		Expdata.ctype = atoi(cv);
		break;
	    case 'd':
		strcpy(Expdata.dec, cv);
		break;
	    case 'e':
		if (mod == 'm')
		    strcpy(Expdata.email, cv);
		else if (mod == 'x')
		    Expdata.x2 = atoi(cv);
		else if (mod == 'y')
		    Expdata.y2 = atoi(cv);
		break;
	    case 'f':
		index = Expdata.flines;
		if (index < 6) {
		    p = &Expdata.fits[index * 80];
		    memset(p, ' ', 80);
		    strcpy(p, cv);
		    Expdata.flines++;
		}
		break;
	    case 'i':
		if (mod == 'n')
		    strcpy(Expdata.inst, cv);
		else
		    strcpy(Expdata.imagetyp, cv);
		break;
	    case 'j':
		strcpy(Expdata.object, cv);
		break;
	    case 'l':
		strcpy(Expdata.lat, cv);
		break;
	    case 'n':
		strcpy(Expdata.longi, cv);
		break;
	    case 'o':
		strcpy(Expdata.obs, cv);
		break;
	    case 'p':
		strcpy(Expdata.psc, cv);
		break;
	    case 'r':
		if (mod == 'a')
		    strcpy(Expdata.ra, cv);
		else {
		    p = Expdata.imagetyp;
		    memset(p, 0, 1560);	// M0002 - I don't like this!
		    Expdata.flines = 0;
		    Expdata.ctype = 0;
		    Expdata.tz = 0;
		}
		break;
	    case 's':
		if (mod == 'u')
		    strcpy(Expdata.subset, cv);
		else if (mod == 'x')
		    Expdata.x1 = atoi(cv);
		else if (mod == 'y')
		    Expdata.y1 = atoi(cv);
		break;
	    case 't':
		strcpy(Expdata.tel, cv);
		break;
	    case 'u':
		strcpy(Expdata.instrum, cv);
		break;
	    case 'w':
		strcpy(Expdata.www, cv);
		break;
	    case 'x':
		strcpy(Expdata.xpix, cv);
		break;
	    case 'y':
		strcpy(Expdata.ypix, cv);
		break;
	    case 'z':
		Expdata.tz = atoi(cv);
		break;
	    }
	} else {
	    // Reached end of ring buffer without seeing a terminating
	    // quote

	    // Dear Lazac, me again! This worries me too. If you reach
	    // the end of the buffer without seeing a terminating quote,
	    // couldn't this just mean that the user has yet to send a
	    // terminating quote? This sounds like we need to return
	    // -ENODATA so that this entire routine can be called
	    // again at a later time when the terminating quote is
	    // available. 

	    // One torture test for this sort of problem is to define a
	    // very small ccd_ring_o_buffer, perhaps 10 bytes long, and
	    // see whether the code falls over! This would be worth
	    // doing.

	    return -ENODATA;
	}
	return 0;
    }
    return -ENODATA;
}

int loop = 0;

spinlock_t ccd_lock = SPIN_LOCK_UNLOCKED;

void ccd_timer_service_task(unsigned long ptr)
{
    assert(ccd_mode.waiting, return);
    ccd_mode.waiting = 0;
    if (ccd_mode.stm) {
	struct timeval now;
	long tl;

	do {
	    do_gettimeofday(&now);
	    tl = (wait_timeval.tv_sec - now.tv_sec) * 1000000L +
		wait_timeval.tv_usec - now.tv_usec;
	    if (tl > 3 * 1000000/HZ ||
		tl < -2 * 1000000/HZ ) {
		printk ("ccd: timing error, %ld usecs\n", tl);
		break;
	    }   
        } while (tl > 0);
    }
    ccd_queue_service_task((int) ptr);
}

void ccd_queue_service_task(int immediate)
{
    unsigned long flags;

    // EXTREME DANGER: what I would like to do here is to use two queues,
    // depending on the urgency of the need, but I'm not sure how to avoid
    // having two copies of ccd_service running.

    if (!module_shutting_down) {
	immediate = 0;

// BEGIN IRQ disabled region ------------------------------------------------

	spin_lock_irqsave(&ccd_lock, flags);
	if (immediate) {
	    queue_task(&ccd_task, &tq_immediate);
	    mark_bh(IMMEDIATE_BH);
	} else {
#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
	    queue_task(&ccd_task, &tq_scheduler);
#else
	    schedule_task(&ccd_task);
#endif
	}
	spin_unlock_irqrestore(&ccd_lock, flags);

// END IRQ disabled region --------------------------------------------------

    }
}

void ccd_service(void *unused)
{
/*
   This is the heart of the device driver, and is called as a task,
   either on the immediate, or scheduler queues. Its function is to
   look after all aspects of the CCD-user interface.  Specifically, it

   (1) Handles port input from the CCD, and places it in the "dma" buffer
       area for reading by "ccd_read()".

   (2) Checks the output ring buffer, and relays any new commands
       to the CCD.

   (3) Wakens reading/writing processes as needed.

   // WARNING: this routine is not reentrant. Therefore, it should only
   // be queued on one task queue at a time.

   // When this routine runs on the immediate queue, the CPU can only
   // handle hardware interrupts.  "schedule ()" can not be called. We
   // therefore should ensure that this routine is very fast, and does
   // not wait for events.

   Michael Ashley / UNSW / 19-Aug-1999 
*/

    int ring_was_full, bytes_consumed;
    int i, j, nchar, ret, n, lo, hi;
    unsigned char c;
    u16 data;
    char *cp;
    struct ccd_command_struct comm;

    if (assertion_has_failed)
	return;

    assert(have_registered_device1 && 
	   have_requested_region && 
	   have_irq &&
	   ccd_setup_to_use_interrupts &&
	   !servicing,
	   return);

    servicing = 1;

    // If we are using port I/O, and some data is expected from the CCD, 
    // try to obtain them.

    if (!ccd_mode.dma_in_progress && ccd_mode.words_expected) {

	// Now loop, obtaining one word at a time, until either the
	// maximum number of words have been obtained, there are no more
	// data immediately available from the CCD, or the "dma" ring buffer
	// becomes full.

	for (i = 0; i < CCD_WORDS_BEFORE_YIELDING; i++) {

	    // DANGER: should loop for 25 microsecs maybe, or do something
	    // clever with FIFO half full.
	    // Perhaps an interrupt can be generated at the end of a row?

#define NASTY_LOOP 1000

	    for (j = 0; j < NASTY_LOOP; j++) {
		if (CCD_IS_SENDING) {
		    break;
		}
	    }
	    if (j == NASTY_LOOP) {
		break;
	    }

	    if (dma_prepare_to_write(ccd_mode.dma_slot, 2)) {
		break;
	    }

	    // We subtract 32768 and swap bytes to conform to the FITS standard.

	    data = inw(CCD_DATA_PORT) - 32768;
	    *((u16 *) dma_write_pointer[ccd_mode.dma_slot]) = htons(data);

	    // Update the write pointer.

	    dma_write_pointer[ccd_mode.dma_slot] =
		dma_next_write_pointer[ccd_mode.dma_slot];

	    // Decrement the number of words expected from the CCD. If
            // none are expected, we can return to reading from the 
            // output ring buffer.

	    if (!ccd_mode.words_expected--) {
		ccd_mode.readout_in_progress = 0;
		break;
	    }
	}
    }

    // Now process the output ring buffer.

    /*
       p{n}  select parameter page n
       V     send a vertical trigger command
       v     stop a vertical transfer
       H     send a horizontal trigger command
       h     stop a horizontal transfer
       .     wait for horizontal transfers
       ,     wait for vertical transfers
       :     stop horizontal transfers
       ;     stop vertical transfers
       W{n}  waits for n milliseconds
       M{n}  wait till the time is zero, modulo n milliseconds
       O     opens the shutter
       C     closes the shutter
       F     turn on the fan
       f     turn off the fan
       T     turn on the peltier cooler
       t     turn off the peltier cooler
       g{n}  set the gain to 0/1
       s{n}  set the data shift (left)
       S{n}  set the data shift (right)
       P{n}  write to the parallel port
       w{n}  write to the parallel port interrupt registers
       z{n}  set the data offset 
       |     select horizontal mode
       -     select vertical mode
       m{n}  set mode n for the currently selected page and direction
       r{n}  set the repeat count
       d{n}  set the vertical delay in microseconds per row
       b{n}  set the horizontal binning factor
       q{n}  0 select port I/O for pixel reads
             1 select DMA for pixel reads
       /{n}  sends CCD output to device special file 0/1
       ^{n}  writes parts of a FITS header, n is used as follows
             0 the first part
             1 the second part
             3 the trailing spaces and END record
       Z{n}  writes n words of zero to the output file
       )     close a FITS file
       I     set FITS header information
       *     a marker used when aborting exposures
       =     initialise the CCD simulator

       e.g:

       :;/0^0
       g0 s0               z00000
       p00 | m0 r0000 d0000 - m4 r0000 b0000
       p01 | m3 r0001 d0100 - m6 r1200 b0000
       p02 | m0 r0000 d0000 - m4 r0800 b0000
       p04 | m3 r1200 d0300 - m5 r0000 b0000
       q1 p2 H V ., p4 V , p2 H V .,
       OW10000C
*/

    ring_was_full = CCD_RING_O_FULL;

    bytes_consumed = 0;	 // The number of bytes we have consumed from the
                         // ccd_ring_o_buffer.

    // EXTREME DANGER: possible infinite loop with 'H' and DMA if the
    // data is not being read, and the DMA buffer fills up.

    // We keep consuming the ccd_ring_o_buffer while it is not empty,
    // while the CCD is not being read out, and while we aren't waiting
    // for a timer to go off.

    while (!CCD_RING_O_EMPTY && !ccd_mode.readout_in_progress &&
	   !ccd_mode.waiting) {

	comm.port = 0;
	comm.delay = 0;
	comm.outword = 0;
	c = ccd_ring_o_buffer[ccd_ring_o_tail];

	// A simple check to avoid an endless loop.

	if (ccd_ring_o_tail != ccd_ring_o_tail_save) {
	    ccd_ring_o_tail_save = ccd_ring_o_tail;
	    loop = 0;
	} else {
	    loop++;
	    assert(loop < 100000, return);
	}

	// nchar is the number of characters read from the
	// ccd_ring_o_buffer during this pass through the loop.

	nchar = 1;

	// ret is the return value from various function calls within
	// the loop. If this goes non-zero, we exit the loop.

	ret = 0;

#if DEBUG_VERBOSE
	printk("%c", c);
#endif

	switch (c) {
	case 'p':		// select parameter page n
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 64, return);
		comm.outword = COMMAND_SELECT_REGISTER + 0x0080 + n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 'V':		// send a vertical trigger command
	    comm.outword = COMMAND_VERTICAL_TRIGGER + 0x0001;
	    ret = ccd_process_command(&comm);
	    break;

	case 'v':		// stop a vertical transfer
	    comm.outword = COMMAND_VERTICAL_TRIGGER;
	    ret = ccd_process_command(&comm);
	    break;

	case 'H':		// send a horizontal trigger command
	    comm.outword = COMMAND_HORIZONTAL_TRIGGER + 0x0001;
	    ret = ccd_process_command(&comm);
	    break;

	case 'h':		// stop a horizontal transfer
	    comm.outword = COMMAND_HORIZONTAL_TRIGGER;
	    ret = ccd_process_command(&comm);
	    break;

	case '.':		// wait for horizontal transfers
	    comm.port = CCD_WAIT_FOR_HORIZONTAL_TRANSFERS;
	    comm.delay = 10000;
	    ret = ccd_process_command(&comm);
	    break;

	case ',':		// wait for vertical transfers
	    comm.port = CCD_WAIT_FOR_VERTICAL_TRANSFERS;
	    comm.delay = 10000;
	    ret = ccd_process_command(&comm);
	    break;

	case ':':		// stop horizontal transfers
	    comm.port = CCD_STOP_HORIZONTAL_TRANSFERS;
	    ret = ccd_process_command(&comm);
	    break;

	case ';':		// stop vertical transfers
	    comm.port = CCD_STOP_VERTICAL_TRANSFERS;
	    ret = ccd_process_command(&comm);
	    break;

	case 'W':	// wait for n milliseconds
	case 'M':	// sleep till the time is zero modulo n milliseconds
	    assert(!ccd_mode.waiting, return);
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		if (!module_shutting_down) {
		    struct timeval now;
		    long now_ms, delay_ms, tl;

                    // Grab the current time in jiffies, and as a timeval.

		    wait_timer.expires = jiffies;
		    do_gettimeofday (&now);

//		    printk ("now %ld %ld %ld\n", wait_timer.expires, now.tv_sec, now.tv_usec);
		    init_timer(&wait_timer);
		    wait_timer.function = ccd_timer_service_task;
		    wait_timer.data = (unsigned long) 1;
		    ccd_mode.waiting = 1;
		    ccd_mode.stm = 1;

                    // Calculate the number of milliseconds we need to delay.

		    if (c == 'M') {
			now_ms = (now.tv_sec % 3600L) * 1000L + now.tv_usec / 1000L;
			delay_ms = n - now_ms % n;
		    } else {
			delay_ms = n;
		    }

                    // Calculate the precise timeval at the end of the delay.

                    tl = delay_ms + (now.tv_usec / 1000L);
		    wait_timeval.tv_sec = now.tv_sec + tl / 1000L;
		    wait_timeval.tv_usec = (tl * 1000L) % 1000000L;

                    // Calculate the jiffies value one tick before the delay.

		    wait_timer.expires += HZ * (delay_ms * 0.001) - 1;
		    add_timer(&wait_timer);
//		    printk ("then %ld %ld %ld %ld\n", wait_timer.expires, wait_timeval.tv_sec, wait_timeval.tv_usec, delay_ms);
		}
	    }
	    break;

	case 'O':		// open the shutter
	    comm.outword = COMMAND_SHUTTER + 0x0001;
	    ret = ccd_process_command(&comm);
	    break;

	case 'C':		// close the shutter
	    comm.outword = COMMAND_SHUTTER;
	    ret = ccd_process_command(&comm);
	    break;

	case 'F':		// turn on the fan
	    comm.outword = COMMAND_COOLER + (ccd_mode.cooler & 0x0002);
	    ret = ccd_process_command(&comm);
	    break;

	case 'f':		// turn off the fan
	    comm.outword = COMMAND_COOLER + (ccd_mode.cooler | 0x0001);
	    ret = ccd_process_command(&comm);
	    break;

	case 'T':		// turn on the peltier cooler
	    comm.outword = COMMAND_COOLER + (ccd_mode.cooler & 0x0001);
	    ret = ccd_process_command(&comm);
	    break;

	case 't':		// turn off the peltier cooler
	    comm.outword = COMMAND_COOLER + (ccd_mode.cooler | 0x0002);
	    ret = ccd_process_command(&comm);
	    break;

	case 'g':		// set the gain to 0/1
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 2, return);
		comm.outword = COMMAND_ANALOG_CONTROL + n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 's':		// set the data shift (left)
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 17, return);
		comm.outword = COMMAND_ANALOG_CONTROL + 0x00c0 + n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 'S':		// set the data shift (right)
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 17, return);
		comm.outword = COMMAND_ANALOG_CONTROL + 0x0080 + n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 'P':		// write to the parallel port
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);
		comm.port = 5;
		comm.outword = n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 'w':		// write to the parallel port interrupt registers
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);
		comm.port = 6;
		comm.outword = n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case 'z':		// set the data offset
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);
		comm.port = 8;
		comm.outword = n;
		ret = ccd_process_command(&comm);
	    }
	    break;

	case '|':		// select horizontal mode
	    ccd_mode.direction = 1;
	    ccd_init.direction = 1;
	    break;

	case '-':		// select vertical mode
	    ccd_mode.direction = 0;
	    ccd_init.direction = 1;
	    break;

	case 'm':		// set mode n for the currently selected page and direction
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(ccd_init.direction, return);
		if (ccd_mode.direction) {
		    assert(n < 16, return);
		    comm.outword = COMMAND_SELECT_REGISTER +
			PARAMETER_VERTICAL_MODE;
		} else {
		    assert(n < 8, return);
		    comm.outword = COMMAND_SELECT_REGISTER +
			PARAMETER_HORIZONTAL_MODE;
		}
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + n;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
	    }
	    break;

	case 'r':		// set the repeat count
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);	// DANGER: artificial limit
		assert(ccd_init.direction, return);
		lo = n % 256;
		hi = n / 256;
		comm.outword = COMMAND_SELECT_REGISTER +
		    (ccd_mode.
		     direction ? PARAMETER_VERTICAL_REPEAT_COUNT_LO :
		     PARAMETER_HORIZONTAL_REPEAT_COUNT_LO);
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + lo;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_SELECT_REGISTER +
		    (ccd_mode.
		     direction ? PARAMETER_VERTICAL_REPEAT_COUNT_HI :
		     PARAMETER_HORIZONTAL_REPEAT_COUNT_HI);
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + hi;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
	    }
	    break;

	case 'd':		// set the vertical delay in microseconds per row
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);
		lo = n % 256;
		hi = n / 256;
		comm.outword = COMMAND_SELECT_REGISTER +
		    PARAMETER_VERTICAL_DELAY_LO;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + lo;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_SELECT_REGISTER +
		    PARAMETER_VERTICAL_DELAY_HI;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + hi;
		if (0 != (ret = ccd_process_command(&comm))) {
		    break;
		}
	    }
	    break;

	case 'b':		// set the horizontal binning factor
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 65536, return);
		lo = n % 256;
		hi = n / 256;
		comm.outword = COMMAND_SELECT_REGISTER +
		    PARAMETER_HORIZONTAL_BINNING_LO;
		if (ccd_process_command(&comm)) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + lo;
		if (ccd_process_command(&comm)) {
		    break;
		}
		comm.outword = COMMAND_SELECT_REGISTER +
		    PARAMETER_HORIZONTAL_BINNING_HI;
		if (ccd_process_command(&comm)) {
		    break;
		}
		comm.outword = COMMAND_WRITE_REGISTER + hi;
		if (ccd_process_command(&comm)) {
		    break;
		}
	    }
	    break;

	case 'q':		// select port I/O (n=1) or DMA (n=2) readout mode
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		switch (n) {
		case 0:
		case 1:
		    ccd_mode.using_dma = n;
		    break;
		default:
		    assert(0, return);
		}
	    }
	    break;

	case '/':		// sends CCD output to device special file 0/1
	    if (ccd_mode.closing_file) {
		ret = -EWOULDBLOCK;
	    } else {
		if ((nchar = ccd_extract_number(&n)) > 0) {
		    assert(n < 2, return);
		    ccd_mode.dma_slot = n;
		}
	    }
	    break;

	case '^':		// writes parts of a FITS header
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n < 4, return);
		if (ccd_mode.fits_index == 0) {
		    ccd_write_fits_header();
		    ccd_mode.fits_index = 1;
		}
		if (0 != (ret = dma_prepare_to_write(0, sizeof(header)))) {
		    break;
		}
		cp = &(header[0][0]);
		for (i = 0; i < sizeof(header); i++) {
		    *((char *) dma_write_pointer[0]++) = *(cp++);
		}
		assert(dma_write_pointer[0] ==
		       dma_next_write_pointer[0], return);
		dma_write_pointer[0] = dma_next_write_pointer[0];
		ccd_mode.fits_index = 0;
	    }
	    break;

	case 'Z':		// writes n words of zero to the output file
	    if ((nchar = ccd_extract_number(&n)) > 0) {
		assert(n > 0, return);
		n *= 2; // convert to bytes
		if (0 != (ret = dma_prepare_to_write(0, n))) {
		    break;
		}
		for (i = 0; i < n; i++) {
		    *((char *) dma_write_pointer[0]++) = 0;
		}
		assert(dma_write_pointer[0] ==
		       dma_next_write_pointer[0], return);
		dma_write_pointer[0] = dma_next_write_pointer[0];
	    }
	    break;

	case ')':		// close a FITS file
	    ccd_mode.closing_file = 1;
	    break;

	case 'I':		// set FITS header information - M0001
	    ret = ccd_parameters(&nchar);
	    break;

	case '=':
#ifdef SIMULATOR
	    ccd_sim_initialise();
#endif
	    comm.outword = COMMAND_STATUS_RESET;
	    ret = ccd_process_command(&comm);
	    break;

	case '\r':
	case '\n':
	case ' ':
	    break;

	default:		// all other characters are considered errors
	
  	    printk("ccd: %ld %ld %ld %d\n", ccd_ring_o_head,
		   ccd_ring_o_tail, ccd_ring_o_tail_save, c);
	    assert(0, (void) 0);
	}

	// EXTREME DANGER: I'm not sure of the ramifications of this, there
	// could be an endless loop.
	// M0001 - YES, it is an endless loop, better putting it after the ring
	// buffer had already been forwarded.
	// M0002 - Actually, I recall thinking that this isn't actually an
	//         endless loop after all, for some subtle reason! I am not 
	//         sure of the ramifications of
	//         moving the "ret < 0" test, but let's leave it at its new
	//         position for now.
	// M0003 - It turns out that this is essential, to cope with
	//        slow commands (such as WAIT_FOR_VERTICAL_TRANSFERS), that
	//        may take some time to complete.

	// We now test for an error in executing the last command (signalled
	// by ret<0), or an error in reading a string of digits (signalled by
	// nchar<0). If an error has occured, we leave this loop without
	// modifying the ring buffer pointer.


	if (ret < 0 || nchar < 0) {
	    break;
	}
	// OK, so the command was executed correctly (or we couldn't decode
	// the associated digit string). We can now update the ring buffer 
	// pointer, and go and get the next command.

	if (nchar > 0) {
	    bytes_consumed += nchar;
	    ccd_ring_o_tail = (ccd_ring_o_tail + nchar) %
		CCD_RING_O_BUFFER_LENGTH;
	}
      

    }

    // If the output ring was previously full, and at least 1 byte has
    // been consumed, then wake up any writing processes.

    if (ring_was_full && bytes_consumed > 0) {
	wake_up_interruptible(&ccd_outq);
    }
    // If either of the DMA buffers has data in it (or, in the case of
    // ccd0, if the file is being closed), wake up any reading
    // processes.

    // NOTE: if there aren't any reading processes, the CCD driver will
    // appear to have crashed. Simply do "cp /dev/ccd0 file" to reawaken.

    if ((dma_read_pointer[0] != dma_write_pointer[0]) ||
	ccd_mode.closing_file) {
	wake_up_interruptible(&ccd_inq0);
    }

    if (dma_read_pointer[1] != dma_write_pointer[1]) {
	wake_up_interruptible(&ccd_inq1);
    }
    // If there is no DMA in progress AND we have not just
    // executed a wait function AND either we were not
    // able to empty the output ring OR there are still words
    // expected from the camera, we reschedule this routine to run
    // the next time the scheduler runs.

    // DANGER: update documentation above.

    if (!ccd_mode.dma_in_progress && !ccd_mode.waiting &&
	ccd_mode.words_expected) {
//	(!CCD_RING_O_EMPTY || ccd_mode.words_expected)) {
	number_of_reschedules++;
#ifdef DEBUG_VERBOSE
	printk(KERN_ALERT "ccd: rescheduling %d\n", number_of_reschedules);
#endif
	ccd_queue_service_task(0);
    }

    servicing = 0;
}

int ccd_initialize(void)
{
/*
   Put hardware initialization code in here.
*/

//----------- At this point, the CCD is initialised. ---------------//

    have_initialized_ccd = 1;
    ccd_setup_to_use_interrupts = 1;	// DANGER: is this needed anymore?

    return 0;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
int ccd_read_procmem(char *buf, char **start, off_t offset,
		     int len, int unused)
#else
int ccd_read_procmem(char *buf, char **start, off_t offset, int len)
#endif
{

    // EXTREME DANGER: how do we avoid overfilling "buf"? On kernels < 2.4,
    // buf appears to be 512 bytes long.

    int page, reg, n, i;
    struct timeval now;
    long tl;
    u16 status;

    assert(ccd_setup_to_use_interrupts, return -EINVAL);

    if (offset > 0) {
	return offset;
    }
    len = 0;
    status = inw_p(CCD_STATUS_PORT);

    len += sprintf(buf + len,
		   "%s device driver, version %s, status 0x%x, %ld\n",
		   ccd_name, ccd_version, status,
		   (long) ccd_mode.time.tv_sec);

    if (ccd_mode.waiting) {
	do_gettimeofday(&now);
	tl = (wait_timeval.tv_sec - now.tv_sec) * 1000000L +
	    wait_timeval.tv_usec - now.tv_usec;
        len += sprintf(buf + len, "Wait countdown = %ld.%02ld secs\n",
		       tl / 1000000, (tl % 1000000) / 10000);
    }

    len += sprintf(buf + len, "Output head-tail = %ld; reschedules = %d\n",
		   ccd_ring_o_head - ccd_ring_o_tail,
		   number_of_reschedules);

    if (ccd_mode.dma_in_progress) {
	len += sprintf(buf + len, "Readout using DMA in progress\n");
    } else if (ccd_mode.readout_in_progress) {
	len += sprintf(buf + len, "Readout using port I/O in progress\n");
    }
    len += sprintf(buf + len, "The DMA buffer has become full %d times\n",
		   ccd_mode.dma_overwrite);

    i = get_dma_residue(dma);
    status = ccd_mode.status;
    len +=
	sprintf(buf + len, "dma residue = %d; status 0x%x\n", i, status);

#ifndef NDEBUG
    for (i = 0; i < 10; i++) {
	len += sprintf(buf + len, "%ld ", dma_times[i]);
    }
#endif

    len += sprintf(buf + len, "Fan is %s, ",
		   ((status & 0x0001) ? "off" : "on"));
    // M0001 - Give status of cooler, too
    i = (status & 0x0006) >> 1;
    len += sprintf(buf + len, "Cooler is %s\n",
		   i == 0 ? "on" :
		   i == 1 ? ">1K below" : i == 2 ? ">1K above" : "off");
    // M0001 - Better to use driver internal flags for the shutter state
    // because the camera status word can be incorrect in case it hasn't
    // been operated since power up.

    len += sprintf(buf + len, "Shutter is %s\n",
		   ((!ccd_init.shutter) ? "unknown" :
		    (ccd_mode.shutter) ? "open" : "closed"));
    len += sprintf(buf + len, "Horizontal operation is%sin progress\n",
		   ((status & 0x0080) ? " " : " not "));
    len += sprintf(buf + len, "Vertical operation is%sin progress\n",
		   ((status & 0x0100) ? " " : " not "));
    len += sprintf(buf + len, "Power supply %s\n",
		   ((status & 0x0400) ? "has had a glitch" : "is OK"));
    len += sprintf(buf + len, "Power supply %s\n",
		   ((status & 0x0800) ? "is currently low" : "is OK"));

#ifndef NDEBUG
    len += sprintf(buf + len, "Last command 0x%x; sequence #%ld\n",
		   last_command, last_command_number);
#endif

    len += sprintf(buf + len,
		   (ccd_init.shutter ?
		    (ccd_mode.shutter ? "O " : "C ") : "  "));
    len += sprintf(buf + len,
		   (ccd_init.cooler ?
		    ((ccd_mode.cooler & 0x0001) ? "f " : "F ") : "  "));
    len += sprintf(buf + len,
		   (ccd_init.cooler ?
		    ((ccd_mode.cooler & 0x0002) ? "t " : "T ") : "  "));
    len += sprintf(buf + len,
		   (ccd_init.gain ?
		    ((ccd_mode.gain & 0x0001) ? "g1 " : "g0 ") : "   "));
    len += sprintf(buf + len,
		   (ccd_init.shift ?
		    ((ccd_mode.
		      shift & 0x0040) ? "s%1d " : "S%1d ") : "   "),
		   ccd_mode.shift & 0x000f);
    len +=
	sprintf(buf + len, (ccd_init.pll ? "I%5.5d " : "       "),
		ccd_mode.pll);
    len +=
	sprintf(buf + len, (ccd_init.pll_int ? "i%5.5d " : "       "),
		ccd_mode.pll_int);
    len +=
	sprintf(buf + len, (ccd_init.offset ? "z%5.5d " : "       "),
		ccd_mode.offset);
    len += sprintf(buf + len, "\n");

    for (page = 0; page < 64; page++) {
	for (reg = 0; reg < 32; reg++) {
	    if (ccd_init.par[page][reg] && ccd_mode.par[page][reg] != 0) {
		len += sprintf(buf + len, "p%2.2d", page);
		if (ccd_init.par[page][PARAMETER_VERTICAL_MODE]) {
		    len += sprintf(buf + len, " | m%1d",
				   ccd_mode.
				   par[page][PARAMETER_VERTICAL_MODE]);
		} else {
		    len += sprintf(buf + len, "     ");
		}
		if (ccd_init.par[page][PARAMETER_VERTICAL_REPEAT_COUNT_LO]
		    || ccd_init.
		    par[page][PARAMETER_VERTICAL_REPEAT_COUNT_HI]) {
		    n = ccd_mode.
			par[page][PARAMETER_VERTICAL_REPEAT_COUNT_LO] +
			ccd_mode.
			par[page][PARAMETER_VERTICAL_REPEAT_COUNT_HI] *
			256;
		    len += sprintf(buf + len, " r%4.4d", n);
		} else {
		    len += sprintf(buf + len, "      ");
		}
		if (ccd_init.par[page][PARAMETER_VERTICAL_DELAY_LO] ||
		    ccd_init.par[page][PARAMETER_VERTICAL_DELAY_HI]) {
		    n = ccd_mode.par[page][PARAMETER_VERTICAL_DELAY_LO] +
			ccd_mode.par[page][PARAMETER_VERTICAL_DELAY_HI] *
			256;
		    len += sprintf(buf + len, " d%4.4d", n);
		} else {
		    len += sprintf(buf + len, "      ");
		}
		if (ccd_init.par[page][PARAMETER_HORIZONTAL_MODE]) {
		    len += sprintf(buf + len, " - m%1d",
				   ccd_mode.
				   par[page][PARAMETER_HORIZONTAL_MODE]);
		} else {
		    len += sprintf(buf + len, "     ");
		}
		if (ccd_init.
		    par[page][PARAMETER_HORIZONTAL_REPEAT_COUNT_LO]
		    || ccd_init.
		    par[page][PARAMETER_HORIZONTAL_REPEAT_COUNT_HI]) {
		    n = ccd_mode.
			par[page][PARAMETER_HORIZONTAL_REPEAT_COUNT_LO] +
			ccd_mode.
			par[page][PARAMETER_HORIZONTAL_REPEAT_COUNT_HI] *
			256;
		    len += sprintf(buf + len, " r%4.4d", n);
		} else {
		    len += sprintf(buf + len, "      ");
		}
		if (ccd_init.par[page][PARAMETER_HORIZONTAL_BINNING_LO] ||
		    ccd_init.par[page][PARAMETER_HORIZONTAL_BINNING_HI]) {
		    n = ccd_mode.
			par[page][PARAMETER_HORIZONTAL_BINNING_LO] +
			ccd_mode.
			par[page][PARAMETER_HORIZONTAL_BINNING_HI] * 256;
		    len += sprintf(buf + len, " b%4.4d", n);
		} else {
		    len += sprintf(buf + len, "      ");
		}
		len += sprintf(buf + len, "\n");
		break;
	    }
	}
    }

    return len;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
struct proc_dir_entry ccd_proc_entry = {
    0,				// low_ino: inode is dynamic
    3, "ccd",			// length of name, and name
    S_IFREG | S_IRUGO,		// mode
    1, 0, 0,			// nlink, uid, gid
    0,				// size (not used)
    NULL,			// operations (use default)
    &ccd_read_procmem,		// function used to provide the data
};
#endif

int open_for_read, open_for_write;

int ccd_open(struct inode *inode, struct file *filp)
{

    if (assertion_has_failed)
	return -EINVAL;

    assert(have_registered_device1 && have_requested_region
	   && have_irq, return -EINVAL);

    if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
	if (open_for_write) {
	    return -EWOULDBLOCK;
	}
	open_for_write = 1;
	filp->private_data = &open_for_write;
    } else {
	if (open_for_read) {
	    return -EWOULDBLOCK;
	}
	open_for_read = 1;
	filp->private_data = &open_for_read;
	
        // DANGER: the following is not ideal. Should check whether we
        // are opening ccd0 or ccd1.

        // We reset the global variable ccd_mode.closing_file, so that
	// other parts of the driver know that the file has been closed,
	// thereby allowing more commands to be sent to the CCD.
		
	ccd_mode.closing_file = 0;
	ccd_queue_service_task(0);
    }

    MOD_INC_USE_COUNT;
    return 0;
}

release_t ccd_release(struct inode * inode, struct file * filp)
{

    assert(have_registered_device1 && have_requested_region
	   && have_irq, release_return(-EINVAL));

    // Remove this filp from the asynchronously notified filp's.

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
    ccd_fasync(inode, filp, 0);
#else
    ccd_fasync(-1, filp, 0);	// DANGER: is "-1" correct?
#endif

    *(int *) (filp->private_data) = 0;

    MOD_DEC_USE_COUNT;
    release_return(0);
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
read_write_t ccd_read0(struct inode *inode, struct file *filp,
		       char *buf, count_t count)
{
#else
read_write_t ccd_read0(struct file *filp,
		       char *buf, size_t count, loff_t * ppos)
{
#endif

    size_t save_count;

    assert(have_registered_device1 && have_requested_region &&
	   have_irq && have_dma, return -EINVAL);

#if LINUX_VERSION_CODE >= VERSION_CODE(2,2,0)
    if (ppos != &filp->f_pos) {
	return -ESPIPE;
    }
#endif

    // Exit the loop if there is data available to be read, or if the
    // file is being closed.

    while (1) {
	if (assertion_has_failed)
	    return -EINVAL;
	
	save_count = count;
	dma_prepare_to_read(0, &count);
	if (count) {
	    break;
	}
	count = save_count;
	
	// EXTREME DANGER: need to think about the ramifications of
	//  .closing_file with respect to port I/O.
	
	if (ccd_mode.closing_file) {
	    count = 0;
	    break;
	}
	// If this was a non-blocking call leave immediately,
	// else sleep waiting for a signal that there is data
	// available (or the file should be closed).
      
	if (filp->f_flags & O_NONBLOCK) {
	    return -EAGAIN;
	}
	cli();
	//printk (" INQ0 ");
	interruptible_sleep_on(&ccd_inq0);
	//printk (" inq0 ");
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
	if (current->signal & ~current->blocked) {
#else
	if (signal_pending(current)) {
#endif
	    sti();
	    return -ERESTARTSYS;
	}
	sti();
    }
	
    // OK, we now have data available for reading, or the file is
    // being closed (in which case, "count" will be zero).  Copy
    // the data to user space, and return the number of bytes
    // copied.
	
    if (count) {
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,4)
	memcpy_tofs(buf, (void *) dma_read_pointer[0], count);
#else
	    
	if (copy_to_user(buf, (void *) dma_read_pointer[0], count)) {
	    return -EFAULT;
	}
#endif
#if DEBUG_VERBOSE
	printk("DMA read: %d bytes at %ld, values %d,%d,%d,%d\n",
	       count, dma_read_pointer[0] - dma_begin[0],
	       *buf, *(buf + 1), *(buf + 2), *(buf + 3));
#endif
	dma_read_pointer[0] = dma_next_read_pointer[0];
	
    } else {
	assert(ccd_mode.closing_file, return -EINVAL);
#if DEBUG_VERBOSE
        printk ("closing file\n");
#endif		
    }
    
#if DEBUG_VERBOSE
    printk ("%d bytes to read\n", count);
#endif		
    
    ccd_queue_service_task(0);
    return count;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
read_write_t ccd_read1(struct inode * inode, struct file * filp,
		       char *buf, count_t count)
{
#else
read_write_t ccd_read1(struct file * filp,
		       char *buf, size_t count, loff_t * ppos)
{
#endif

    size_t save_count;

    assert(have_registered_device1 && have_requested_region &&
	   have_irq && have_dma, return -EINVAL);

#if LINUX_VERSION_CODE >= VERSION_CODE(2,2,0)
    if (ppos != &filp->f_pos) {
	return -ESPIPE;
    }
#endif

    // Exit the loop if there is data available to be read.

    while (1) {
	if (assertion_has_failed)
	    return -EINVAL;

	save_count = count;
	dma_prepare_to_read(1, &count);
	if (count) {
	    break;
	}
	count = save_count;

	// If this was a non-blocking call leave immediately,
	// else sleep waiting for a signal that there is data
	// available.

	if (filp->f_flags & O_NONBLOCK) {
	    return -EAGAIN;
	}
	cli();
	interruptible_sleep_on(&ccd_inq1);
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
	if (current->signal & ~current->blocked) {
#else
	if (signal_pending(current)) {
#endif
	    sti();
	    return -ERESTARTSYS;
	}
	sti();
    }

    // OK, we now have data available for reading.
    // Copy the data to user space, and return the number of bytes copied.

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,4)
    memcpy_tofs(buf, (void *) dma_read_pointer[1], count);
#else
    if (copy_to_user(buf, (void *) dma_read_pointer[1], count)) {
	return -EFAULT;
    }
#endif

    dma_read_pointer[1] = dma_next_read_pointer[1];

    ccd_queue_service_task(0);
    return count;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
read_write_t ccd_write(struct inode * inode, struct file * filp,
		       const char *buf, count_t count)
{
#else
read_write_t ccd_write(struct file * filp,
		       const char *buf, size_t count, loff_t * ppos)
{
#endif

    unsigned char local_buf[CCD_WORDS_BEFORE_YIELDING];
    count_t i;

    assert(have_registered_device1 && have_requested_region &&
	   have_irq && have_dma, return -EINVAL);

#if LINUX_VERSION_CODE >= VERSION_CODE(2,2,0)
    if (ppos != &filp->f_pos) {
	return -ESPIPE;
    }
#endif

    while (CCD_RING_O_FULL) {
	if (assertion_has_failed)
	    return -EINVAL;
	if (filp->f_flags & O_NONBLOCK) {
	    return -EAGAIN;
	}
	cli();			// To prevent a race condition (see "ccd_select" for more info).

	if (CCD_RING_O_FULL) {
	    interruptible_sleep_on(&ccd_outq);
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
	    if (current->signal & ~current->blocked) {
#else
	    if (signal_pending(current)) {
#endif
		sti();
		return -ERESTARTSYS;
	    }
	}
	sti();
    }

    // At this point, we have room for at least one byte in the output
    // ring buffer.

    if (count > sizeof local_buf) {
	count = sizeof local_buf;
    }
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,4)
    memcpy_fromfs(local_buf, buf, count);
#else
    if (copy_from_user(local_buf, buf, count)) {
	printk("yike!\n");	// DANGER

	return -EFAULT;
    }
#endif

    for (i = 0; i < count; i++) {
	ccd_ring_o_buffer[ccd_ring_o_head] = local_buf[i];
	ccd_ring_o_head = (ccd_ring_o_head + 1) % CCD_RING_O_BUFFER_LENGTH;
	if (CCD_RING_O_FULL) {
	    i++;
	    break;
	}
    }

    ccd_queue_service_task(0);

    // EXTREME DANGER: is it OK to have the ccd_task scheduled on two
    // queues at the same time?

    return i;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
int ccd_fsync(struct inode *inode, struct file *filp)
{
#elif LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
int ccd_fsync(struct file *filp, struct dentry *d)
{
#else
int ccd_fsync(struct file *filp, struct dentry *d, int datasync)
{
#endif

/* 
   Guarantees that all buffered data has been sent to the CCD.
   If "datasync" is non-zero, then this function should act like
   fdatasync (which is the same as fsync, except that we do not have
   to update metadata such as access time). In this implementation, we
   do not distinguish between fsync and fdatasync.

   Michael Ashley / UNSW / 09-Jan-1999
 */

    // EXTREME DANGER: this isn't right yet, we probably need another wait
    // queue entry, to be triggered when the output ring is empty?
    // Alternatively, do we really need an fsync?

    while (!CCD_RING_O_EMPTY) {
	if (assertion_has_failed)
	    return -EINVAL;
	cli();			// To prevent a race condition (see "ccd_select" for more info).

	if (!CCD_RING_O_EMPTY) {
	    interruptible_sleep_on(&ccd_outq);
	}
	sti();
    }
    return (0);
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
int ccd_fasync(struct inode *inode, struct file *filp, int mode)
{

/* 
   Michael Ashley / UNSW / 26-Mar-1999
 */

    return fasync_helper(inode, filp, mode, &ccd_async_queue);
}
#else
int ccd_fasync(int fd, struct file *filp, int mode)
{

/* 
   Michael Ashley / UNSW / 26-Mar-1999
 */

    return fasync_helper(fd, filp, mode, &ccd_async_queue);
}
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,23)

int ccd_select0(struct inode *inode, struct file *filp,
		int mode, select_table * table)
{

    if (assertion_has_failed)
	return -EINVAL;

    assert(have_registered_device1 && have_requested_region &&
	   have_irq && have_dma, return -EINVAL);

    if (mode == SEL_IN) {

	cli();
	// This is to prevent a race condition before going to
	// sleep in the "select_wait ()" call. Note that
	// we don't need to "save_flags ()" since the
	// operating system guarantees that interrupts are
	// enabled before entering device driver methods.

	if (dma_read_pointer[0] == dma_write_pointer[0]) {
	    select_wait(&ccd_inq0, table);
	    sti();
	    return 0;
	} else {
	    sti();
	    return 1;
	}
    }
    if (mode == SEL_OUT) {
	cli();
	if (CCD_RING_O_FULL) {
	    select_wait(&ccd_outq, table);
	    sti();
	    return 0;
	} else {
	    sti();
	    return 1;
	}
    }
    // M0001 - If we get here we should return an error.
    // SEL_EX can never happen with a Wright camera, because it can happen
    // in only two cases:
    // 1. when out-of-band data arrives for a socket (we are not a socket)
    // 2. presence of control status info to be read from the master side
    //    of a pseudo terminal that has been put into packet mode (and we
    //    are not a pseudo terminal either)

    // M0002 - I put a hook for an exception in, as a mechanism to cope
    // with hardware errors. For example, suppose the camera doesn't
    // respond as it should (e.g., a command times out), then perhaps we
    // could send an exception (a bit like out-of-band data) to let a
    // monitoring process know that something has gone wrong. The camera
    // could then be power-cycled, and the driver reloaded.

    if (mode == SEL_EX) {

    }

    return -EINVAL;
}
#else

unsigned int ccd_poll0(struct file *filp, poll_table * table)
{

    unsigned int result = 0;

    if (assertion_has_failed)
	return -EINVAL;

    assert(have_registered_device1 && have_requested_region
	   && have_irq, return -EINVAL);

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,89)
    poll_wait(&ccd_inq0, table);
    poll_wait(&ccd_outq, table);
    poll_wait(&ccd_exq, table);
#else
    poll_wait(filp, &ccd_inq0, table);
    poll_wait(filp, &ccd_outq, table);
    poll_wait(filp, &ccd_exq, table);
#endif

    cli();
    // DANGER: I'm not sure if this is necessary, see comments in
    //         ccd_select above.

    if (dma_read_pointer[0] != dma_write_pointer[0]) {
	result = result || POLLIN || POLLRDNORM;
    }
    if (!CCD_RING_O_FULL) {
	result = result || POLLOUT || POLLWRNORM;
    }
    sti();

    return (result);
}
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,23)

int ccd_select1(struct inode *inode, struct file *filp,
		int mode, select_table * table)
{

    if (assertion_has_failed)
	return -EINVAL;

    assert(have_registered_device1 && have_requested_region &&
	   have_irq && have_dma, return -EINVAL);

    if (mode == SEL_IN) {

	cli();
	// This is to prevent a race condition before going to
	// sleep in the "select_wait ()" call. Note that
	// we don't need to "save_flags ()" since the
	// operating system guarantees that interrupts are
	// enabled before entering device driver methods.

	if (dma_read_pointer[1] == dma_write_pointer[1]) {
	    select_wait(&ccd_inq1, table);
	    sti();
	    return 0;
	} else {
	    sti();
	    return 1;
	}
    }
    if (mode == SEL_OUT || mode == SEL_EX) {
	return 0;
    }
    assert(0, return -EINVAL);	// Control should never reach here.

    return (0);
}
#else

unsigned int ccd_poll1(struct file *filp, poll_table * table)
{

    unsigned int result = 0;

    if (assertion_has_failed)
	return -EINVAL;

    assert(have_registered_device1 && have_requested_region
	   && have_irq, return -EINVAL);

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,89)
    poll_wait(&ccd_inq1, table);
#else
    poll_wait(filp, &ccd_inq1, table);
#endif

    cli();
    // DANGER: I'm not sure if this is necessary, see comments in
    //         ccd_select above.

    if (dma_read_pointer[1] != dma_write_pointer[1]) {
	result = result || POLLIN || POLLRDNORM;
    }
    sti();

    return (result);
}
#endif

#define CCD_IOC_MAGIC   'v'
#define CCD_IOCSABORT   _IO(CCD_IOC_MAGIC, 1)	// M0001
#define CCD_IOCSCANCEL  _IO(CCD_IOC_MAGIC, 2)	// M0001
#define CCD_IOCSHANGUP  _IO(CCD_IOC_MAGIC, 3)	// M0001
#define CCD_IOCSRESUME  _IO(CCD_IOC_MAGIC, 4)	// M0001
#define CCD_IOCSTEST    _IO(CCD_IOC_MAGIC, 5)	// M0004
#define CCD_IOC_MAXNR   5	// M0001

int ccd_ioctl(struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg)
{

// Only IOCSABORT and IOCSTEST are currently implemented.

// If this routine is called, the user must have opened the device driver
// for write access, which means that no one else can have write access.

    int i,err = 0, size;
    unsigned long flags;

    if (assertion_has_failed)
	return -EINVAL;

    size = _IOC_SIZE(cmd);

    if (_IOC_TYPE(cmd) != CCD_IOC_MAGIC)
	return -EINVAL;
    if (_IOC_NR(cmd) > CCD_IOC_MAXNR)
	return -EINVAL;
    if (_IOC_DIR(cmd) & _IOC_READ) {
	err = verify_area(VERIFY_WRITE, (void *) arg, size);
    } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
	err = verify_area(VERIFY_READ, (void *) arg, size);
    }
    if (err)
	return err;

    switch (cmd) {

    case CCD_IOCSABORT:

	printk("ccd: aborting exposure\n");

// BEGIN IRQ disabled region ------------------------------------------------

	spin_lock_irqsave(&ccd_lock, flags);

        // Prevents the ISR from adding more data.

        if (ccd_mode.dma_in_progress) {
	    throw_away_dma_data = 1;
	}

	// Triggers the closing of the output file, if necessary.

	if (open_for_read) {
	    ccd_mode.closing_file = 1;
	}

        // Zero the ccd_ring_o_buffer.

	ccd_ring_o_head = 0;
	ccd_ring_o_tail = 0;
	ccd_ring_o_tail_save = 0;

	// Zero the DMA buffers.

	dma_allocate_buffer(0);

        // Remove any exposure timers.

	if (ccd_mode.waiting) {
	    del_timer(&wait_timer);
	    ccd_mode.waiting = 0;
	} 
	spin_unlock_irqrestore(&ccd_lock, flags);

// END IRQ disabled region --------------------------------------------------

        // Queue the service task, to process the file closure and wake up
        // reading processes.
	
	ccd_queue_service_task(0);
	break;

    case CCD_IOCSCANCEL:
	printk("ccd: cancelling exposure\n");
	break;

    case CCD_IOCSHANGUP:
	printk("ccd: hangup exposure\n");
	break;

    case CCD_IOCSRESUME:
	break;

    case CCD_IOCSTEST:
        printk ("ccd: test mode - recommend reloading the module\n");
	for (i = 0; i < 1000; i++) {
	    ccd_send_command (COMMAND_STATUS_RETURN);
	}
	break;

    default:
	assert(0, return -EINVAL);
    }
    return 0;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
lseek_t ccd_llseek(struct inode * inode, struct file * filp,
		   off_t off, int whence)
{
#else
lseek_t ccd_llseek(struct file * filp, loff_t off, int whence)
{
#endif
    return -ESPIPE;		// Unseekable.

}

static int ccd_find_hardware(void)
{				/* Returns zero for success, and
				   a negative number on error. */

    int result;
    int port;
    int i;
    unsigned int ret;

    assert(have_registered_device1 && !have_requested_region
	   && !have_irq, return -EINVAL);

    result = 0;

    if (io != 0) {
	ccd_possible_port_bases[0] = io;
	ccd_possible_port_bases[1] = 0;
    }
    printk(KERN_NOTICE "ccd: / module loading, version %s\n", ccd_version);
    printk(KERN_NOTICE "ccd: | probing for io base...trying ");

// EXTREME DANGER: shouldn't do this here. this is for testing only.

    if (irq == 0)		// M0001
	irq = CCD_IRQ_NUMBER;

    have_irq = 0 == request_irq(irq, ccd_isr,
				SA_INTERRUPT || SA_SAMPLE_RANDOM,
				"ccd", NULL);
    if (!have_irq) {
	printk(KERN_INFO "ccd: | can't get assigned irq %i\n", irq);
	return -ENODEV;
    }
// END EXTREME DANGER

    port = 0;

    for (i = 0; ccd_possible_port_bases[i]; i++) {
	if (port != 0) {
	    break;
	}
	io = ccd_possible_port_bases[i];
	if (!check_region(io, CCD_IO_EXTENT)) {
	    printk("0x%x, ", io);

	    /* We now do something to check whether there is a CCD at "io". */

	    /* Read a word from the interface status port. */

	    ret = inw_p(CCD_STATUS_PORT);

	    /* If the "transmit ready" bit 7 is high, and the 
	       "always 0" bit 1 is low,
	       then we may have found the camera. We investigate further... */

	    printk ("-%x-",ret);
	    if ((ret & 0x80) && !(ret & 0x02)) {

		/* Write some random data to the interface status port, and then read
		   the port again. */

		outw_p(0x005a, CCD_STATUS_PORT);
		ret = inw_p(CCD_STATUS_PORT);

		/* If the port is like a memory location, we will read the same data
		   that we wrote, so this isn't a camera. If the data is different, and
		   the "transmit ready" bit is high, we assume that we have found the
		   camera. */
		printk (".");

		if ((ret & 0x80) && (ret != 0x005a)) {

		    /* DANGER: let's assume that the camera is number 0 */

		    ccd_number = 0 * 0x1000;
                    inw_p(CCD_DATA_PORT); // clear any data remaining to be read
		    outw_p(0x0000, CCD_DMA_PORT); // disable DMA
		    outw_p(0x0000, CCD_CAM_INT_PORT); // disable interrupts
		    outw_p(0x0000, CCD_CANCEL_CAM_INT_PORT); // cancel camera interrupt
		    outw_p(0x0000, CCD_PLL_INT_PORT); // disable interrupts
		    outw_p(0x0000, CCD_CANCEL_PLL_INT_PORT); // cancel interrupts
		    if (0 != ccd_send_command (COMMAND_STATUS_RESET)) {
			printk ("\nccd: unable to reset status word\nccd: ");
		    }

#if 0
		    /* Iterate over all possible camera numbers at this base location. */

		    for (i = 0; i <= 7; i++) {
			ccd_number = i * 0x1000;

			/* Assuming there is a camera with this number, stop all
			   transfers, and then ask it to send back one data
			   word. If a word is sent back, then we have found a
			   camera. */

			status = 0;

			// DANGER: here and below, for the CCD
#if 0
			STOP_VERTICAL_TRANSFERS(status);
			STOP_HORIZONTAL_TRANSFERS(status);
			SELECT_PAGE(0, status);
			SET_HORIZONTAL_MODE(HighSpeed_Data_WhenTriggered,
					    status);
			SET_HORIZONTAL_REPEAT_COUNT(1, status);
			SET_HORIZONTAL_BINNING(1, status);
			HORIZONTAL_TRIGGER(status);
#endif
			/* Give the camera a couple of milliseconds to respond. */

			udelay(2000);

			/* Is there a word waiting to be read? */

			if (CCD_IS_SENDING) {

			    /* We appear to have found the camera number, so reset it. */
#if 0
			    CLOSE_SHUTTER(status);
			    FAN_AND_COOLER_ON(status);
			    RESET_STATUS_WORD(status);
#endif
#endif
			    //          if (status == 0) {
			    port = io;
			    //}
#if 0
			    break;
			}
		    }
#endif
		}
	    }
	} else {
	    printk("skipping 0x%x, ", io);
	}
    }

    if (port == 0) {
	printk("%s not found\n", ccd_name);
	return -ENODEV;
    } else {
	printk("%s found\n", ccd_name);
    }

    request_region(io, CCD_IO_EXTENT, "ccd");
    have_requested_region = 1;

    result = ccd_initialize();

    if (result < 0) {
        printk(KERN_CRIT "ccd: | initialization failure\n");
	return result;
    }
// We now get the IRQ...

// EXTREME DANGER: should disable the camera interrupt before
    // registering an interrupt handler.
#if 0
    irq = CCD_IRQ_NUMBER;

    have_irq = 0 == request_irq(irq, ccd_isr,
				SA_INTERRUPT || SA_SAMPLE_RANDOM,
				"ccd", NULL);
    if (!have_irq) {
	printk(KERN_INFO "ccd: | can't get assigned irq %i\n", irq);
	return -ENODEV;
    }
#endif

    // and the DMA channel. Begin by requesting the channel from the
    // operating system.

    if (dma == 0)		// M0001
	dma = CCD_DMA_CHANNEL;

    have_dma = 0 == request_dma(dma, "ccd");

    if (!have_dma) {
	printk(KERN_INFO "ccd: | can't get assigned dma %i\n", dma);
	return -ENODEV;
    }
    ret = dma_allocate_buffer(MAX_WORDS_PER_DMA_TRANSFER * 16);
    if (ret) {
	printk(KERN_INFO "ccd: | can't allocate memory for DMA\n");
	return ret;
    }

    printk(KERN_INFO "ccd: | probing for IRQ, found at %i\n", irq);


    printk(KERN_NOTICE "ccd: | %s successfully initialized\n", ccd_name);

    return 0;
}

void ccd_cleanup(unsigned long ptr)
{
    wake_up(&cleanup_queue);
}

void cleanup_module(void)
{
    printk(KERN_NOTICE "ccd: \\ module unloading, ");

    // I'm not at all sure how to guarantee that all tasks associated
    // with this module have stopped. So we use a brute force technique
    // of setting a global variable module_shutting_down, and then waiting
    // for a certain number of seconds. This should ensure that any
    // pending DMA is complete, and that all tasks have run to completion.

    throw_away_dma_data = 1;
    module_shutting_down = 1;
    del_timer(&wait_timer);

    // M0001 - Sleep for SECS_TO_WAIT_BEFORE_CLEANUP seconds.
    // There has got to be an easier way to do this. What happened to
    // the current->timeout technique? (e.g., Rubini p136).

    init_timer(&cleanup_timer);
    cleanup_timer.function = ccd_cleanup;
    cleanup_timer.data = (unsigned long) 0;
    cleanup_timer.expires = jiffies + HZ * SECS_TO_WAIT_BEFORE_CLEANUP;
    add_timer(&cleanup_timer);
    sleep_on(&cleanup_queue);

#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,0)
    flush_scheduled_tasks();
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
    proc_unregister(&proc_root, ccd_proc_entry.low_ino);
#else
    remove_proc_entry("ccd", NULL);
#endif

    if (have_requested_region) {

	// Stop the CCD from causing interrupts. Note that we can only do
	// this if we have requested an IO region (else we don't know what
	// the CCD IO base address is).

	// Release the IO address registers.

	release_region(io, CCD_IO_EXTENT);
	printk("io region, ");
    }

    // Release the DMA channel.

    if (have_dma) {
	free_dma(dma);
	printk("dma, ");
    }

    // Release the PC's IRQ.

    if (have_irq) {
	free_irq(irq, NULL);
	printk("irq, ");
    }

    // Release the major device number.

    if (have_registered_device0) {
	unregister_chrdev(ccd_major0, "ccd0");
    }
    if (have_registered_device1) {
	unregister_chrdev(ccd_major1, "ccd1");
    }
    printk("shutdown complete\n");
}

int init_module(void)
{
    int result;

    ccd_task.routine = ccd_service;

#if LINUX_VERSION_CODE >= VERSION_CODE(2,4,0)
    init_waitqueue_head(&cleanup_queue);
    init_waitqueue_head(&ccd_inq0);
    init_waitqueue_head(&ccd_inq1);
    init_waitqueue_head(&ccd_outq);
    init_waitqueue_head(&ccd_exq);
#endif

    // Attempt to register the major node numbers for the two device
    // special files required by the driver. If "ccd{0,1}_major" is zero, 
    // we obtain dynamically allocated major node numbers.

    result = register_chrdev(ccd_major0, "ccd0", &ccd_fops0);
    if (result < 0) {
	printk(KERN_WARNING "ccd: | can't get major number %d\n",
	       ccd_major0);
	cleanup_module();
	return result;
    }
    if (ccd_major0 == 0)
	ccd_major0 = result;
    have_registered_device0 = 1;

    result = register_chrdev(ccd_major1, "ccd1", &ccd_fops1);
    if (result < 0) {
	printk(KERN_WARNING "ccd: | can't get major number %d\n",
	       ccd_major1);
	cleanup_module();
	return result;
    }
    if (ccd_major1 == 0)
	ccd_major1 = result;
    have_registered_device1 = 1;

    // Now try to identify the CCD's io base address and hardware IRQ, and
    // then initialize the CCD.

#ifdef SIMULATOR
    ccd_sim_initialise();
#endif				/* SIMULATOR */

    result = ccd_find_hardware();
    if (result < 0) {
	cleanup_module();
	return result;
    }
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
    proc_register_dynamic(&proc_root, &ccd_proc_entry);
#elif LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
    proc_register(&proc_root, &ccd_proc_entry);
#else
    create_proc_info_entry("ccd", 0, NULL, ccd_read_procmem);
#endif

    // M0001 - Initialize our FITS header container
    // Set up an initial state for shutter
    memset(&Expdata, 0, sizeof(Expdata));

    // Success!

    return 0;
}

// Nonzero if YEAR is a leap year (every 4 years,
// except every 100th isn't, and every 400th is).
#define __isleap(year)  \
     ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

// How many days come before each month (0-12).
unsigned short int __mon_yday[2][13] = {
    /* Normal years.  */
    {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365},
    /* Leap years.  */
    {0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366}
};

#define SECS_PER_HOUR   (60 * 60)
#define SECS_PER_DAY    (SECS_PER_HOUR * 24)

// Compute the `struct tm' representation of *t,
// offset OFFSET seconds east of UTC,
// and store year, yday, mon, mday, wday, hour, min, sec into *TP.
void offtime(time_t * t, long int offset, struct tm *tp)
{
    long int days, rem, y;
    unsigned short int *ip;

    days = *t / SECS_PER_DAY;
    rem = *t % SECS_PER_DAY;
    rem += offset;
    while (rem < 0) {
	rem += SECS_PER_DAY;
	--days;
    }
    while (rem >= SECS_PER_DAY) {
	rem -= SECS_PER_DAY;
	++days;
    }
    tp->tm_hour = rem / SECS_PER_HOUR;
    rem %= SECS_PER_HOUR;
    tp->tm_min = rem / 60;
    tp->tm_sec = rem % 60;
    /* January 1, 1970 was a Thursday.  */
    tp->tm_wday = (4 + days) % 7;
    if (tp->tm_wday < 0)
	tp->tm_wday += 7;
    y = 1970;

#define LEAPS_THRU_END_OF(y) ((y) / 4 - (y) / 100 + (y) / 400)

    while (days < 0 || days >= (__isleap(y) ? 366 : 365)) {
	/* Guess a corrected year, assuming 365 days per year.  */
	long int yg = y + days / 365 - (days % 365 < 0);

	/* Adjust DAYS and Y to match the guessed year.  */
	days -= ((yg - y) * 365 + LEAPS_THRU_END_OF(yg - 1)
		 - LEAPS_THRU_END_OF(y - 1));
	y = yg;
    }
    tp->tm_year = y - 1900;
    tp->tm_yday = days;
    ip = __mon_yday[__isleap(y)];
    for (y = 11; days < ip[y]; --y)
	continue;
    days -= ip[y];
    tp->tm_mon = y;
    tp->tm_mday = days + 1;
}

MODULE_AUTHOR("Michael Ashley <m.ashley@unsw.edu.au>");
MODULE_DESCRIPTION("Wright Instruments CCD camera driver");
